import { ContainerInterface, ContainerOptions, ContainerModuleInterface, AsyncContainerModuleInterface, ServiceIdentifier, BindingToSyntax as BindingToSyntaxInterface, Middleware, Newable, BindingActivation, BindingDeactivation } from '../interfaces/interfaces';
import { MetadataReader } from '../planning/metadata_reader';
import { BindingToSyntax } from '../syntax/binding_to_syntax';
export declare class Container implements ContainerInterface {
    id: number;
    parent: ContainerInterface | null;
    readonly options: ContainerOptions;
    private _middleware;
    private _bindingDictionary;
    private _activations;
    private _deactivations;
    private _snapshots;
    private _metadataReader;
    private _moduleActivationStore;
    static merge(container1: ContainerInterface, container2: ContainerInterface, ...containers: ContainerInterface[]): ContainerInterface;
    constructor(containerOptions?: ContainerOptions);
    load(...modules: ContainerModuleInterface[]): void;
    loadAsync(...modules: AsyncContainerModuleInterface[]): Promise<void>;
    unload(...modules: ContainerModuleInterface[]): void;
    bind<T>(serviceIdentifier: ServiceIdentifier<T>): BindingToSyntaxInterface<T>;
    rebind<T>(serviceIdentifier: ServiceIdentifier<T>): BindingToSyntaxInterface<T>;
    rebindAsync<T>(serviceIdentifier: ServiceIdentifier<T>): Promise<BindingToSyntax<T>>;
    unbind(serviceIdentifier: ServiceIdentifier): void;
    unbindAsync(serviceIdentifier: ServiceIdentifier): Promise<void>;
    unbindAll(): void;
    unbindAllAsync(): Promise<void>;
    onActivation<T>(serviceIdentifier: ServiceIdentifier<T>, onActivation: BindingActivation<T>): void;
    onDeactivation<T>(serviceIdentifier: ServiceIdentifier<T>, onDeactivation: BindingDeactivation<T>): void;
    isBound(serviceIdentifier: ServiceIdentifier<any>): boolean;
    isBoundNamed(serviceIdentifier: ServiceIdentifier<any>, named: string | number | symbol): boolean;
    isBoundTagged(serviceIdentifier: ServiceIdentifier<any>, key: string | number | symbol, value: any): boolean;
    snapshot(): void;
    restore(): void;
    createChild(containerOptions?: ContainerOptions): Container;
    applyMiddleware(...middlewares: Middleware[]): void;
    applyCustomMetadataReader(metadataReader: MetadataReader): void;
    get<T>(serviceIdentifier: ServiceIdentifier<T>): T;
    getAsync<T>(serviceIdentifier: ServiceIdentifier<T>): Promise<T>;
    getTagged<T>(serviceIdentifier: ServiceIdentifier<T>, key: string | number | symbol, value: any): T;
    getTaggedAsync<T>(serviceIdentifier: ServiceIdentifier<T>, key: string | number | symbol, value: unknown): Promise<T>;
    getNamed<T>(serviceIdentifier: ServiceIdentifier<T>, named: string | number | symbol): T;
    getNamedAsync<T>(serviceIdentifier: ServiceIdentifier<T>, named: string | number | symbol): Promise<T>;
    getAll<T>(serviceIdentifier: ServiceIdentifier<T>): T[];
    getAllAsync<T>(serviceIdentifier: ServiceIdentifier<T>): Promise<T[]>;
    getAllTagged<T>(serviceIdentifier: ServiceIdentifier<T>, key: string | number | symbol, value: any): T[];
    getAllTaggedAsync<T>(serviceIdentifier: ServiceIdentifier<T>, key: string | number | symbol, value: unknown): Promise<T[]>;
    getAllNamed<T>(serviceIdentifier: ServiceIdentifier<T>, named: string | number | symbol): T[];
    getAllNamedAsync<T>(serviceIdentifier: ServiceIdentifier<T>, named: string | number | symbol): Promise<T[]>;
    resolve<T>(constructorFunction: Newable<T>): T;
    private _preDestroy;
    private _removeModuleHandlers;
    private _removeModuleBindings;
    private _deactivate;
    private _handleDeactivationError;
    private _deactivateContainer;
    private _deactivateContainerAsync;
    private _getContainerModuleHelpersFactory;
    private _getAll;
    private _get;
    private _getButThrowIfAsync;
    private _getAllArgs;
    private _getNotAllArgs;
    private _planAndResolve;
    private _deactivateIfSingleton;
    private _deactivateSingletons;
    private _deactivateSingletonsAsync;
    private _propagateContainerDeactivationThenBindingAndPreDestroy;
    private _propagateContainerDeactivationThenBindingAndPreDestroyAsync;
    private _removeServiceFromDictionary;
    private _bindingDeactivationAndPreDestroy;
    private _bindingDeactivationAndPreDestroyAsync;
}
