'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var reflection = require('@abraham/reflection');

/**
 * Used for named bindings
 */
var NAMED_TAG = 'named';
/**
 * The name of the target at design time
 */

var NAME_TAG = 'name';
/**
 * The for unmanaged injections (in base classes when using inheritance)
 */

var UNMANAGED_TAG = 'unmanaged';
/**
 * The for optional injections
 */

var OPTIONAL_TAG = 'optional';
/**
 * The type of the binding at design time
 */

var INJECT_TAG = 'inject';
/**
 * The type of the binding at design type for multi-injections
 */

var MULTI_INJECT_TAG = 'multi_inject';
/**
 * used to store constructor arguments tags
 */

var TAGGED = 'inversify:tagged';
/**
 * used to store class properties tags
 */

var TAGGED_PROP = 'inversify:tagged_props';
/**
 * used to store types to be injected
 */

var PARAM_TYPES = 'inversify:paramtypes'; // used to access design time types

var DESIGN_PARAM_TYPES = 'design:paramtypes'; // used to identify postConstruct functions

var POST_CONSTRUCT = 'post_construct'; // used to identify preDestroy functions

var PRE_DESTROY = 'pre_destroy';

function getNonCustomTagKeys() {
  return [INJECT_TAG, MULTI_INJECT_TAG, NAME_TAG, UNMANAGED_TAG, NAMED_TAG, OPTIONAL_TAG];
}

var NON_CUSTOM_TAG_KEYS = /*#__PURE__*/getNonCustomTagKeys();

var metadata_keys = {
  __proto__: null,
  NAMED_TAG: NAMED_TAG,
  NAME_TAG: NAME_TAG,
  UNMANAGED_TAG: UNMANAGED_TAG,
  OPTIONAL_TAG: OPTIONAL_TAG,
  INJECT_TAG: INJECT_TAG,
  MULTI_INJECT_TAG: MULTI_INJECT_TAG,
  TAGGED: TAGGED,
  TAGGED_PROP: TAGGED_PROP,
  PARAM_TYPES: PARAM_TYPES,
  DESIGN_PARAM_TYPES: DESIGN_PARAM_TYPES,
  POST_CONSTRUCT: POST_CONSTRUCT,
  PRE_DESTROY: PRE_DESTROY,
  NON_CUSTOM_TAG_KEYS: NON_CUSTOM_TAG_KEYS
};

var BindingScopeEnum = {
  Request: 'Request',
  Singleton: 'Singleton',
  Transient: 'Transient'
};
var BindingTypeEnum = {
  ConstantValue: 'ConstantValue',
  Constructor: 'Constructor',
  DynamicValue: 'DynamicValue',
  Factory: 'Factory',
  Function: 'Function',
  Instance: 'Instance',
  Invalid: 'Invalid',
  Provider: 'Provider'
};
var TargetTypeEnum = {
  ClassProperty: 'ClassProperty',
  ConstructorArgument: 'ConstructorArgument',
  Variable: 'Variable'
};

function _regeneratorRuntime() {
  /*! regenerator-runtime -- Copyright (c) 2014-present, Facebook, Inc. -- license (MIT): https://github.com/facebook/regenerator/blob/main/LICENSE */

  _regeneratorRuntime = function () {
    return exports;
  };

  var exports = {},
      Op = Object.prototype,
      hasOwn = Op.hasOwnProperty,
      $Symbol = "function" == typeof Symbol ? Symbol : {},
      iteratorSymbol = $Symbol.iterator || "@@iterator",
      asyncIteratorSymbol = $Symbol.asyncIterator || "@@asyncIterator",
      toStringTagSymbol = $Symbol.toStringTag || "@@toStringTag";

  function define(obj, key, value) {
    return Object.defineProperty(obj, key, {
      value: value,
      enumerable: !0,
      configurable: !0,
      writable: !0
    }), obj[key];
  }

  try {
    define({}, "");
  } catch (err) {
    define = function (obj, key, value) {
      return obj[key] = value;
    };
  }

  function wrap(innerFn, outerFn, self, tryLocsList) {
    var protoGenerator = outerFn && outerFn.prototype instanceof Generator ? outerFn : Generator,
        generator = Object.create(protoGenerator.prototype),
        context = new Context(tryLocsList || []);
    return generator._invoke = function (innerFn, self, context) {
      var state = "suspendedStart";
      return function (method, arg) {
        if ("executing" === state) throw new Error("Generator is already running");

        if ("completed" === state) {
          if ("throw" === method) throw arg;
          return doneResult();
        }

        for (context.method = method, context.arg = arg;;) {
          var delegate = context.delegate;

          if (delegate) {
            var delegateResult = maybeInvokeDelegate(delegate, context);

            if (delegateResult) {
              if (delegateResult === ContinueSentinel) continue;
              return delegateResult;
            }
          }

          if ("next" === context.method) context.sent = context._sent = context.arg;else if ("throw" === context.method) {
            if ("suspendedStart" === state) throw state = "completed", context.arg;
            context.dispatchException(context.arg);
          } else "return" === context.method && context.abrupt("return", context.arg);
          state = "executing";
          var record = tryCatch(innerFn, self, context);

          if ("normal" === record.type) {
            if (state = context.done ? "completed" : "suspendedYield", record.arg === ContinueSentinel) continue;
            return {
              value: record.arg,
              done: context.done
            };
          }

          "throw" === record.type && (state = "completed", context.method = "throw", context.arg = record.arg);
        }
      };
    }(innerFn, self, context), generator;
  }

  function tryCatch(fn, obj, arg) {
    try {
      return {
        type: "normal",
        arg: fn.call(obj, arg)
      };
    } catch (err) {
      return {
        type: "throw",
        arg: err
      };
    }
  }

  exports.wrap = wrap;
  var ContinueSentinel = {};

  function Generator() {}

  function GeneratorFunction() {}

  function GeneratorFunctionPrototype() {}

  var IteratorPrototype = {};
  define(IteratorPrototype, iteratorSymbol, function () {
    return this;
  });
  var getProto = Object.getPrototypeOf,
      NativeIteratorPrototype = getProto && getProto(getProto(values([])));
  NativeIteratorPrototype && NativeIteratorPrototype !== Op && hasOwn.call(NativeIteratorPrototype, iteratorSymbol) && (IteratorPrototype = NativeIteratorPrototype);
  var Gp = GeneratorFunctionPrototype.prototype = Generator.prototype = Object.create(IteratorPrototype);

  function defineIteratorMethods(prototype) {
    ["next", "throw", "return"].forEach(function (method) {
      define(prototype, method, function (arg) {
        return this._invoke(method, arg);
      });
    });
  }

  function AsyncIterator(generator, PromiseImpl) {
    function invoke(method, arg, resolve, reject) {
      var record = tryCatch(generator[method], generator, arg);

      if ("throw" !== record.type) {
        var result = record.arg,
            value = result.value;
        return value && "object" == typeof value && hasOwn.call(value, "__await") ? PromiseImpl.resolve(value.__await).then(function (value) {
          invoke("next", value, resolve, reject);
        }, function (err) {
          invoke("throw", err, resolve, reject);
        }) : PromiseImpl.resolve(value).then(function (unwrapped) {
          result.value = unwrapped, resolve(result);
        }, function (error) {
          return invoke("throw", error, resolve, reject);
        });
      }

      reject(record.arg);
    }

    var previousPromise;

    this._invoke = function (method, arg) {
      function callInvokeWithMethodAndArg() {
        return new PromiseImpl(function (resolve, reject) {
          invoke(method, arg, resolve, reject);
        });
      }

      return previousPromise = previousPromise ? previousPromise.then(callInvokeWithMethodAndArg, callInvokeWithMethodAndArg) : callInvokeWithMethodAndArg();
    };
  }

  function maybeInvokeDelegate(delegate, context) {
    var method = delegate.iterator[context.method];

    if (undefined === method) {
      if (context.delegate = null, "throw" === context.method) {
        if (delegate.iterator.return && (context.method = "return", context.arg = undefined, maybeInvokeDelegate(delegate, context), "throw" === context.method)) return ContinueSentinel;
        context.method = "throw", context.arg = new TypeError("The iterator does not provide a 'throw' method");
      }

      return ContinueSentinel;
    }

    var record = tryCatch(method, delegate.iterator, context.arg);
    if ("throw" === record.type) return context.method = "throw", context.arg = record.arg, context.delegate = null, ContinueSentinel;
    var info = record.arg;
    return info ? info.done ? (context[delegate.resultName] = info.value, context.next = delegate.nextLoc, "return" !== context.method && (context.method = "next", context.arg = undefined), context.delegate = null, ContinueSentinel) : info : (context.method = "throw", context.arg = new TypeError("iterator result is not an object"), context.delegate = null, ContinueSentinel);
  }

  function pushTryEntry(locs) {
    var entry = {
      tryLoc: locs[0]
    };
    1 in locs && (entry.catchLoc = locs[1]), 2 in locs && (entry.finallyLoc = locs[2], entry.afterLoc = locs[3]), this.tryEntries.push(entry);
  }

  function resetTryEntry(entry) {
    var record = entry.completion || {};
    record.type = "normal", delete record.arg, entry.completion = record;
  }

  function Context(tryLocsList) {
    this.tryEntries = [{
      tryLoc: "root"
    }], tryLocsList.forEach(pushTryEntry, this), this.reset(!0);
  }

  function values(iterable) {
    if (iterable) {
      var iteratorMethod = iterable[iteratorSymbol];
      if (iteratorMethod) return iteratorMethod.call(iterable);
      if ("function" == typeof iterable.next) return iterable;

      if (!isNaN(iterable.length)) {
        var i = -1,
            next = function next() {
          for (; ++i < iterable.length;) if (hasOwn.call(iterable, i)) return next.value = iterable[i], next.done = !1, next;

          return next.value = undefined, next.done = !0, next;
        };

        return next.next = next;
      }
    }

    return {
      next: doneResult
    };
  }

  function doneResult() {
    return {
      value: undefined,
      done: !0
    };
  }

  return GeneratorFunction.prototype = GeneratorFunctionPrototype, define(Gp, "constructor", GeneratorFunctionPrototype), define(GeneratorFunctionPrototype, "constructor", GeneratorFunction), GeneratorFunction.displayName = define(GeneratorFunctionPrototype, toStringTagSymbol, "GeneratorFunction"), exports.isGeneratorFunction = function (genFun) {
    var ctor = "function" == typeof genFun && genFun.constructor;
    return !!ctor && (ctor === GeneratorFunction || "GeneratorFunction" === (ctor.displayName || ctor.name));
  }, exports.mark = function (genFun) {
    return Object.setPrototypeOf ? Object.setPrototypeOf(genFun, GeneratorFunctionPrototype) : (genFun.__proto__ = GeneratorFunctionPrototype, define(genFun, toStringTagSymbol, "GeneratorFunction")), genFun.prototype = Object.create(Gp), genFun;
  }, exports.awrap = function (arg) {
    return {
      __await: arg
    };
  }, defineIteratorMethods(AsyncIterator.prototype), define(AsyncIterator.prototype, asyncIteratorSymbol, function () {
    return this;
  }), exports.AsyncIterator = AsyncIterator, exports.async = function (innerFn, outerFn, self, tryLocsList, PromiseImpl) {
    void 0 === PromiseImpl && (PromiseImpl = Promise);
    var iter = new AsyncIterator(wrap(innerFn, outerFn, self, tryLocsList), PromiseImpl);
    return exports.isGeneratorFunction(outerFn) ? iter : iter.next().then(function (result) {
      return result.done ? result.value : iter.next();
    });
  }, defineIteratorMethods(Gp), define(Gp, toStringTagSymbol, "Generator"), define(Gp, iteratorSymbol, function () {
    return this;
  }), define(Gp, "toString", function () {
    return "[object Generator]";
  }), exports.keys = function (object) {
    var keys = [];

    for (var key in object) keys.push(key);

    return keys.reverse(), function next() {
      for (; keys.length;) {
        var key = keys.pop();
        if (key in object) return next.value = key, next.done = !1, next;
      }

      return next.done = !0, next;
    };
  }, exports.values = values, Context.prototype = {
    constructor: Context,
    reset: function (skipTempReset) {
      if (this.prev = 0, this.next = 0, this.sent = this._sent = undefined, this.done = !1, this.delegate = null, this.method = "next", this.arg = undefined, this.tryEntries.forEach(resetTryEntry), !skipTempReset) for (var name in this) "t" === name.charAt(0) && hasOwn.call(this, name) && !isNaN(+name.slice(1)) && (this[name] = undefined);
    },
    stop: function () {
      this.done = !0;
      var rootRecord = this.tryEntries[0].completion;
      if ("throw" === rootRecord.type) throw rootRecord.arg;
      return this.rval;
    },
    dispatchException: function (exception) {
      if (this.done) throw exception;
      var context = this;

      function handle(loc, caught) {
        return record.type = "throw", record.arg = exception, context.next = loc, caught && (context.method = "next", context.arg = undefined), !!caught;
      }

      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i],
            record = entry.completion;
        if ("root" === entry.tryLoc) return handle("end");

        if (entry.tryLoc <= this.prev) {
          var hasCatch = hasOwn.call(entry, "catchLoc"),
              hasFinally = hasOwn.call(entry, "finallyLoc");

          if (hasCatch && hasFinally) {
            if (this.prev < entry.catchLoc) return handle(entry.catchLoc, !0);
            if (this.prev < entry.finallyLoc) return handle(entry.finallyLoc);
          } else if (hasCatch) {
            if (this.prev < entry.catchLoc) return handle(entry.catchLoc, !0);
          } else {
            if (!hasFinally) throw new Error("try statement without catch or finally");
            if (this.prev < entry.finallyLoc) return handle(entry.finallyLoc);
          }
        }
      }
    },
    abrupt: function (type, arg) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];

        if (entry.tryLoc <= this.prev && hasOwn.call(entry, "finallyLoc") && this.prev < entry.finallyLoc) {
          var finallyEntry = entry;
          break;
        }
      }

      finallyEntry && ("break" === type || "continue" === type) && finallyEntry.tryLoc <= arg && arg <= finallyEntry.finallyLoc && (finallyEntry = null);
      var record = finallyEntry ? finallyEntry.completion : {};
      return record.type = type, record.arg = arg, finallyEntry ? (this.method = "next", this.next = finallyEntry.finallyLoc, ContinueSentinel) : this.complete(record);
    },
    complete: function (record, afterLoc) {
      if ("throw" === record.type) throw record.arg;
      return "break" === record.type || "continue" === record.type ? this.next = record.arg : "return" === record.type ? (this.rval = this.arg = record.arg, this.method = "return", this.next = "end") : "normal" === record.type && afterLoc && (this.next = afterLoc), ContinueSentinel;
    },
    finish: function (finallyLoc) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];
        if (entry.finallyLoc === finallyLoc) return this.complete(entry.completion, entry.afterLoc), resetTryEntry(entry), ContinueSentinel;
      }
    },
    catch: function (tryLoc) {
      for (var i = this.tryEntries.length - 1; i >= 0; --i) {
        var entry = this.tryEntries[i];

        if (entry.tryLoc === tryLoc) {
          var record = entry.completion;

          if ("throw" === record.type) {
            var thrown = record.arg;
            resetTryEntry(entry);
          }

          return thrown;
        }
      }

      throw new Error("illegal catch attempt");
    },
    delegateYield: function (iterable, resultName, nextLoc) {
      return this.delegate = {
        iterator: values(iterable),
        resultName: resultName,
        nextLoc: nextLoc
      }, "next" === this.method && (this.arg = undefined), ContinueSentinel;
    }
  }, exports;
}

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };
  return _extends.apply(this, arguments);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };
  return _setPrototypeOf(o, p);
}

function _isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Boolean.prototype.valueOf.call(Reflect.construct(Boolean, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _construct(Parent, args, Class) {
  if (_isNativeReflectConstruct()) {
    _construct = Reflect.construct.bind();
  } else {
    _construct = function _construct(Parent, args, Class) {
      var a = [null];
      a.push.apply(a, args);
      var Constructor = Function.bind.apply(Parent, a);
      var instance = new Constructor();
      if (Class) _setPrototypeOf(instance, Class.prototype);
      return instance;
    };
  }

  return _construct.apply(null, arguments);
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _createForOfIteratorHelperLoose(o, allowArrayLike) {
  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];
  if (it) return (it = it.call(o)).next.bind(it);

  if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
    if (it) o = it;
    var i = 0;
    return function () {
      if (i >= o.length) return {
        done: true
      };
      return {
        done: false,
        value: o[i++]
      };
    };
  }

  throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var idCounter = 0;
function id() {
  return idCounter++;
}

var Binding = /*#__PURE__*/function () {
  /**
   * Determines weather the bindings has been already activated
   * The activation action takes place when an instance is resolved
   * If the scope is singleton it only happens once
   */

  /**
   * A runtime identifier because at runtime we don't have interfaces
   */

  /**
   * The constructor of a class which must implement T
   */

  /**
   * Cache used to allow singleton scope and BindingType.ConstantValue bindings
   */

  /**
   * Cache used to allow BindingType.DynamicValue bindings
   */

  /**
   * The scope mode to be used
   */

  /**
   * The kind of binding
   */

  /**
   * A factory method used in BindingType.Factory bindings
   */

  /**
   * An async factory method used in BindingType.Provider bindings
   */

  /**
   * A constraint used to limit the contexts in which this binding is applicable
   */

  /**
   * On activation handler (invoked just before an instance is added to cache and injected)
   */
  // On deactivation handler (invoked just before an instance is unbinded and removed from container)
  function Binding(serviceIdentifier, scope) {
    this.id = void 0;
    this.moduleId = void 0;
    this.activated = void 0;
    this.serviceIdentifier = void 0;
    this.implementationType = void 0;
    this.cache = void 0;
    this.dynamicValue = void 0;
    this.scope = void 0;
    this.type = void 0;
    this.factory = void 0;
    this.provider = void 0;
    this.constraint = void 0;
    this.onActivation = void 0;
    this.onDeactivation = void 0;
    this.id = id();
    this.activated = false;
    this.serviceIdentifier = serviceIdentifier;
    this.scope = scope;
    this.type = BindingTypeEnum.Invalid; // @ts-ignore

    this.constraint = function (_request) {
      return true;
    };

    this.implementationType = null;
    this.cache = null;
    this.factory = null;
    this.provider = null;
    this.onActivation = null;
    this.onDeactivation = null;
    this.dynamicValue = null;
  }

  var _proto = Binding.prototype;

  _proto.clone = function clone() {
    var clone = new Binding(this.serviceIdentifier, this.scope);
    clone.activated = clone.scope === BindingScopeEnum.Singleton ? this.activated : false;
    clone.implementationType = this.implementationType;
    clone.dynamicValue = this.dynamicValue;
    clone.scope = this.scope;
    clone.type = this.type;
    clone.factory = this.factory;
    clone.provider = this.provider;
    clone.constraint = this.constraint;
    clone.onActivation = this.onActivation;
    clone.onDeactivation = this.onDeactivation;
    clone.cache = this.cache;
    return clone;
  };

  return Binding;
}();

var DUPLICATED_INJECTABLE_DECORATOR = 'Cannot apply @injectable decorator multiple times.';
var DUPLICATED_METADATA = 'Metadata key was used more than once in a parameter:';
var NULL_ARGUMENT = 'NULL argument';
var KEY_NOT_FOUND = 'Key Not Found';
var AMBIGUOUS_MATCH = 'Ambiguous match found for serviceIdentifier:';
var CANNOT_UNBIND = 'Could not unbind serviceIdentifier:';
var NOT_REGISTERED = 'No matching bindings found for serviceIdentifier:';
var MISSING_INJECTABLE_ANNOTATION = 'Missing required @injectable annotation in:';
var MISSING_INJECT_ANNOTATION = 'Missing required @inject or @multiInject annotation in:';
var UNDEFINED_INJECT_ANNOTATION = function UNDEFINED_INJECT_ANNOTATION(name) {
  return "@inject called with undefined this could mean that the class " + name + " has " + 'a circular dependency problem. You can use a LazyServiceIdentifer to  ' + 'overcome this limitation.';
};
var CIRCULAR_DEPENDENCY = 'Circular dependency found:';
var INVALID_BINDING_TYPE = 'Invalid binding type:';
var NO_MORE_SNAPSHOTS_AVAILABLE = 'No snapshot available to restore.';
var INVALID_MIDDLEWARE_RETURN = 'Invalid return type in middleware. Middleware must return!';
var INVALID_FUNCTION_BINDING = 'Value provided to function binding must be a function!';
var LAZY_IN_SYNC = function LAZY_IN_SYNC(key) {
  return "You are attempting to construct '" + key + "' in a synchronous way\n but it has asynchronous dependencies.";
};
var INVALID_TO_SELF_VALUE = 'The toSelf function can only be applied when a constructor is ' + 'used as service identifier';
var INVALID_DECORATOR_OPERATION = 'The @inject @multiInject @tagged and @named decorators ' + 'must be applied to the parameters of a class constructor or a class property.';
var ARGUMENTS_LENGTH_MISMATCH = function ARGUMENTS_LENGTH_MISMATCH() {
  return 'The number of constructor arguments in the derived class ' + ((arguments.length <= 0 ? undefined : arguments[0]) + " must be >= than the number of constructor arguments of its base class.");
};
var CONTAINER_OPTIONS_MUST_BE_AN_OBJECT = 'Invalid Container constructor argument. Container options ' + 'must be an object.';
var CONTAINER_OPTIONS_INVALID_DEFAULT_SCOPE = 'Invalid Container option. Default scope must ' + "be a string ('singleton' or 'transient').";
var CONTAINER_OPTIONS_INVALID_AUTO_BIND_INJECTABLE = 'Invalid Container option. Auto bind injectable must be a boolean';
var CONTAINER_OPTIONS_INVALID_SKIP_BASE_CHECK = 'Invalid Container option. Skip base check must be a boolean';
var MULTIPLE_POST_CONSTRUCT_METHODS = 'Cannot apply @postConstruct decorator multiple times in the same class';
var ASYNC_UNBIND_REQUIRED = 'Attempting to unbind dependency with asynchronous destruction (@preDestroy or onDeactivation)';
var POST_CONSTRUCT_ERROR = function POST_CONSTRUCT_ERROR() {
  return "@postConstruct error in class " + (arguments.length <= 0 ? undefined : arguments[0]) + ": " + (arguments.length <= 1 ? undefined : arguments[1]);
};
var PRE_DESTROY_ERROR = function PRE_DESTROY_ERROR(clazz, errorMessage) {
  return "@preDestroy error in class " + clazz + ": " + errorMessage;
};
var ON_DEACTIVATION_ERROR = function ON_DEACTIVATION_ERROR(clazz, errorMessage) {
  return "onDeactivation() error in class " + clazz + ": " + errorMessage;
};
var CIRCULAR_DEPENDENCY_IN_FACTORY = function CIRCULAR_DEPENDENCY_IN_FACTORY(factoryType, serviceIdentifier) {
  return "It looks like there is a circular dependency in one of the '" + factoryType + "' bindings. Please investigate bindings with" + ("service identifier '" + serviceIdentifier + "'.");
};
var STACK_OVERFLOW = 'Maximum call stack size exceeded';

var MetadataReader = /*#__PURE__*/function () {
  function MetadataReader() {}

  var _proto = MetadataReader.prototype;

  _proto.getConstructorMetadata = function getConstructorMetadata(constructorFunc) {
    // TypeScript compiler generated annotations
    var compilerGeneratedMetadata = reflection.getMetadata(PARAM_TYPES, constructorFunc); // User generated constructor annotations

    var userGeneratedMetadata = reflection.getMetadata(TAGGED, constructorFunc);
    return {
      compilerGeneratedMetadata: compilerGeneratedMetadata,
      userGeneratedMetadata: userGeneratedMetadata || {}
    };
  };

  _proto.getPropertiesMetadata = function getPropertiesMetadata(constructorFunc) {
    // User generated properties annotations
    var userGeneratedMetadata = reflection.getMetadata(TAGGED_PROP, constructorFunc) || [];
    return userGeneratedMetadata;
  };

  return MetadataReader;
}();

var BindingCount = {
  MultipleBindingsAvailable: 2,
  NoBindingsAvailable: 0,
  OnlyOneBindingAvailable: 1
};

function isStackOverflowExeption(error) {
  return error instanceof RangeError || error.message === STACK_OVERFLOW;
}
var tryAndThrowErrorIfStackOverflow = function tryAndThrowErrorIfStackOverflow(fn, errorCallback) {
  try {
    return fn();
  } catch (error) {
    if (isStackOverflowExeption(error)) {
      error = errorCallback(); // eslint-disable-line
    }

    throw error;
  }
};

function getServiceIdentifierAsString(serviceIdentifier) {
  if (typeof serviceIdentifier === 'function') {
    var _serviceIdentifier = serviceIdentifier;
    return _serviceIdentifier.name;
  } else if (typeof serviceIdentifier === 'symbol') {
    return serviceIdentifier.toString();
  } else {
    var _serviceIdentifier2$t;

    // string
    var _serviceIdentifier2 = serviceIdentifier;
    return (_serviceIdentifier2$t = _serviceIdentifier2 == null ? void 0 : _serviceIdentifier2.toString()) != null ? _serviceIdentifier2$t : _serviceIdentifier2;
  }
}
function listRegisteredBindingsForServiceIdentifier(container, serviceIdentifier, getBindings) {
  var registeredBindingsList = '';
  var registeredBindings = getBindings(container, serviceIdentifier);

  if (registeredBindings.length !== 0) {
    registeredBindingsList = '\nRegistered bindings:';
    registeredBindings.forEach(function (binding) {
      // Use "Object as name of constant value injections"
      var name = 'Object'; // Use function name if available

      if (binding.implementationType !== null) {
        name = getFunctionName(binding.implementationType);
      }

      registeredBindingsList = registeredBindingsList + "\n " + name;

      if (binding.constraint.metaData) {
        registeredBindingsList = registeredBindingsList + " - " + binding.constraint.metaData;
      }
    });
  }

  return registeredBindingsList;
}

function alreadyDependencyChain(request, serviceIdentifier) {
  if (request.parentRequest === null) {
    return false;
  } else if (request.parentRequest.serviceIdentifier === serviceIdentifier) {
    return true;
  } else {
    return alreadyDependencyChain(request.parentRequest, serviceIdentifier);
  }
}

function dependencyChainToString(request) {
  function _createStringArr(req, result) {
    if (result === void 0) {
      result = [];
    }

    var serviceIdentifier = getServiceIdentifierAsString(req.serviceIdentifier);
    result.push(serviceIdentifier);

    if (req.parentRequest !== null) {
      return _createStringArr(req.parentRequest, result);
    }

    return result;
  }

  var stringArr = _createStringArr(request);

  return stringArr.reverse().join(' --> ');
}

function circularDependencyToException(request) {
  request.childRequests.forEach(function (childRequest) {
    if (alreadyDependencyChain(childRequest, childRequest.serviceIdentifier)) {
      var services = dependencyChainToString(childRequest);
      throw new Error(CIRCULAR_DEPENDENCY + " " + services);
    } else {
      circularDependencyToException(childRequest);
    }
  });
}
function listMetadataForTarget(serviceIdentifierString, target) {
  if (target.isTagged() || target.isNamed()) {
    var m = '';
    var namedTag = target.getNamedTag();
    var otherTags = target.getCustomTags();

    if (namedTag !== null) {
      m += namedTag.toString() + '\n';
    }

    if (otherTags !== null) {
      otherTags.forEach(function (tag) {
        m += tag.toString() + '\n';
      });
    }

    return " " + serviceIdentifierString + "\n " + serviceIdentifierString + " - " + m;
  } else {
    return " " + serviceIdentifierString;
  }
}
function getFunctionName(v) {
  if (v.name) {
    return v.name;
  } else {
    var name = v.toString();
    var match = name.match(/^function\s*([^\s(]+)/);
    return match ? match[1] : "Anonymous function: " + name;
  }
}
function getSymbolDescription(symbol) {
  return symbol.toString().slice(7, -1);
}

var Context = /*#__PURE__*/function () {
  function Context(container) {
    this.id = void 0;
    this.container = void 0;
    this.plan = void 0;
    this.currentRequest = void 0;
    this.id = id();
    this.container = container;
  }

  var _proto = Context.prototype;

  _proto.addPlan = function addPlan(plan) {
    this.plan = plan;
  };

  _proto.setCurrentRequest = function setCurrentRequest(currentRequest) {
    this.currentRequest = currentRequest;
  };

  return Context;
}();

var Metadata = /*#__PURE__*/function () {
  function Metadata(key, value) {
    this.key = void 0;
    this.value = void 0;
    this.key = key;
    this.value = value;
  }

  var _proto = Metadata.prototype;

  _proto.toString = function toString() {
    if (this.key === NAMED_TAG) {
      return "named: " + this.value.toString() + " ";
    } else {
      return "tagged: { key:" + this.key.toString() + ", value: " + this.value + " }";
    }
  };

  return Metadata;
}();

var Plan = function Plan(parentContext, rootRequest) {
  this.parentContext = void 0;
  this.rootRequest = void 0;
  this.parentContext = parentContext;
  this.rootRequest = rootRequest;
};

function injectable() {
  return function (target) {
    var _getMetadata;

    if (reflection.hasOwnMetadata(PARAM_TYPES, target)) {
      throw new Error(DUPLICATED_INJECTABLE_DECORATOR);
    }

    var types = (_getMetadata = reflection.getMetadata(DESIGN_PARAM_TYPES, target)) != null ? _getMetadata : [];
    reflection.defineMetadata(PARAM_TYPES, types, target);
    return target;
  };
}

function isPromise(object) {
  var isObjectOrFunction = typeof object === 'object' && object !== null || typeof object === 'function';
  return isObjectOrFunction && typeof object.then === 'function';
}
function isPromiseOrContainsPromise(object) {
  if (isPromise(object)) {
    return true;
  }

  return Array.isArray(object) && object.some(isPromise);
}

exports.FactoryType = void 0;

(function (FactoryType) {
  FactoryType["DynamicValue"] = "toDynamicValue";
  FactoryType["Factory"] = "toFactory";
  FactoryType["Provider"] = "toProvider";
})(exports.FactoryType || (exports.FactoryType = {}));

var multiBindToService = function multiBindToService(container) {
  return function (service) {
    return function () {
      for (var _len = arguments.length, types = new Array(_len), _key = 0; _key < _len; _key++) {
        types[_key] = arguments[_key];
      }

      return types.forEach(function (t) {
        return container.bind(t).toService(service);
      });
    };
  };
};
var ensureFullyBound = function ensureFullyBound(binding) {
  var boundValue = null;

  switch (binding.type) {
    case BindingTypeEnum.ConstantValue:
    case BindingTypeEnum.Function:
      boundValue = binding.cache;
      break;

    case BindingTypeEnum.Constructor:
    case BindingTypeEnum.Instance:
      boundValue = binding.implementationType;
      break;

    case BindingTypeEnum.DynamicValue:
      boundValue = binding.dynamicValue;
      break;

    case BindingTypeEnum.Provider:
      boundValue = binding.provider;
      break;

    case BindingTypeEnum.Factory:
      boundValue = binding.factory;
      break;
  }

  if (boundValue === null) {
    // The user probably created a binding but didn't finish it
    // e.g. container.bind<T>("Something"); missing BindingToSyntax
    var serviceIdentifierAsString = getServiceIdentifierAsString(binding.serviceIdentifier);
    throw new Error(INVALID_BINDING_TYPE + " " + serviceIdentifierAsString);
  }
};
var getFactoryDetails = function getFactoryDetails(binding) {
  switch (binding.type) {
    case BindingTypeEnum.Factory:
      return {
        factory: binding.factory,
        factoryType: exports.FactoryType.Factory
      };

    case BindingTypeEnum.Provider:
      return {
        factory: binding.provider,
        factoryType: exports.FactoryType.Provider
      };

    case BindingTypeEnum.DynamicValue:
      return {
        factory: binding.dynamicValue,
        factoryType: exports.FactoryType.DynamicValue
      };

    default:
      throw new Error("Unexpected factory type " + binding.type);
  }
};

function isClonable(obj) {
  return typeof obj === 'object' && obj !== null && 'clone' in obj && typeof obj.clone === 'function';
}

function getFirstArrayDuplicate(array) {
  var seenValues = new Set();

  for (var _iterator = _createForOfIteratorHelperLoose(array), _step; !(_step = _iterator()).done;) {
    var entry = _step.value;

    if (seenValues.has(entry)) {
      return entry;
    } else {
      seenValues.add(entry);
    }
  }

  return undefined;
}

function targetIsConstructorFunction(target) {
  return target.prototype !== undefined;
}

function _throwIfMethodParameter(parameterName) {
  if (parameterName !== undefined) {
    throw new Error(INVALID_DECORATOR_OPERATION);
  }
}

function tagParameter(annotationTarget, parameterName, parameterIndex, metadata) {
  _throwIfMethodParameter(parameterName);

  _tagParameterOrProperty(TAGGED, annotationTarget, parameterIndex.toString(), metadata);
}
function tagProperty(annotationTarget, propertyName, metadata) {
  if (targetIsConstructorFunction(annotationTarget)) {
    throw new Error(INVALID_DECORATOR_OPERATION);
  }

  _tagParameterOrProperty(TAGGED_PROP, annotationTarget.constructor, propertyName, metadata);
}

function _ensureNoMetadataKeyDuplicates(metadata) {
  var metadatas = [];

  if (Array.isArray(metadata)) {
    metadatas = metadata;
    var duplicate = getFirstArrayDuplicate(metadatas.map(function (md) {
      return md.key;
    }));

    if (duplicate !== undefined) {
      throw new Error(DUPLICATED_METADATA + " " + duplicate.toString());
    }
  } else {
    metadatas = [metadata];
  }

  return metadatas;
}

function _tagParameterOrProperty(metadataKey, annotationTarget, key, metadata) {
  var _paramOrPropertyMetad;

  var metadatas = _ensureNoMetadataKeyDuplicates(metadata);

  var paramsOrPropertiesMetadata = {}; // read metadata if available

  if (reflection.hasOwnMetadata(metadataKey, annotationTarget)) {
    paramsOrPropertiesMetadata = reflection.getMetadata(metadataKey, annotationTarget);
  } // get metadata for the decorated parameter by its index


  var paramOrPropertyMetadata = paramsOrPropertiesMetadata[key];

  if (paramOrPropertyMetadata === undefined) {
    paramOrPropertyMetadata = [];
  } else {
    var _loop = function _loop() {
      var m = _step.value;

      if (metadatas.some(function (md) {
        return md.key === m.key;
      })) {
        throw new Error(DUPLICATED_METADATA + " " + m.key.toString());
      }
    };

    for (var _iterator = _createForOfIteratorHelperLoose(paramOrPropertyMetadata), _step; !(_step = _iterator()).done;) {
      _loop();
    }
  } // set metadata


  (_paramOrPropertyMetad = paramOrPropertyMetadata).push.apply(_paramOrPropertyMetad, metadatas);

  paramsOrPropertiesMetadata[key] = paramOrPropertyMetadata;
  reflection.defineMetadata(metadataKey, paramsOrPropertiesMetadata, annotationTarget);
}

function createTaggedDecorator(metadata) {
  return function (target, targetKey, indexOrPropertyDescriptor) {
    if (typeof indexOrPropertyDescriptor === 'number') {
      tagParameter(target, targetKey, indexOrPropertyDescriptor, metadata);
    } else {
      tagProperty(target, targetKey, metadata);
    }
  };
}

function _decorate(decorators, target) {
  reflection.decorate(decorators, target);
}

function _param(paramIndex, decorator) {
  return function (target, key) {
    decorator(target, key, paramIndex);
  };
}
/**
 * Allows VanillaJS developers to use decorators:
 *
 * @export
 * @param {((ClassDecorator | ParameterDecorator | MethodDecorator))} decorator
 * @param {*} target
 * @param {(number | string)} [parameterIndex]
 * @example
 * decorate(injectable("Foo", "Bar"), FooBar);
 * decorate(targetName("foo", "bar"), FooBar);
 * decorate(named("foo"), FooBar, 0);
 * decorate(tagged("bar"), FooBar, 1);
 */


function decorate(decorator, target, parameterIndex) {
  if (typeof parameterIndex === 'number') {
    _decorate([_param(parameterIndex, decorator)], target);
  } else if (typeof parameterIndex === 'string') {
    reflection.decorate([decorator], target, parameterIndex);
  } else {
    _decorate([decorator], target);
  }
}

// @ts-ignore

function tagged(metadataKey, metadataValue) {
  return createTaggedDecorator(new Metadata(metadataKey, metadataValue));
}

/**
 * Used to add named metadata which is used to resolve name-based contextual bindings.
 *
 * @export
 * @param {(string | number | symbol)} name
 * @return {*}
 */

function named(name) {
  return createTaggedDecorator(new Metadata(NAMED_TAG, name));
}

function injectBase(metadataKey) {
  return function (serviceIdentifier) {
    return function (target, targetKey, indexOrPropertyDescriptor) {
      if (serviceIdentifier === undefined) {
        var className = typeof target === 'function' ? target.name : target.constructor.name;
        throw new Error(UNDEFINED_INJECT_ANNOTATION(className));
      }

      return createTaggedDecorator(new Metadata(metadataKey, serviceIdentifier))(target, targetKey, indexOrPropertyDescriptor);
    };
  };
}

var inject = /*#__PURE__*/injectBase(INJECT_TAG);

var LazyServiceIdentifer = /*#__PURE__*/function () {
  function LazyServiceIdentifer(cb) {
    this._cb = void 0;
    this._cb = cb;
  }

  var _proto = LazyServiceIdentifer.prototype;

  _proto.unwrap = function unwrap() {
    return this._cb();
  };

  return LazyServiceIdentifer;
}();

function optional() {
  return createTaggedDecorator(new Metadata(OPTIONAL_TAG, true));
}

function unmanaged() {
  return function (target, targetKey, index) {
    var metadata = new Metadata(UNMANAGED_TAG, true);
    tagParameter(target, targetKey, index, metadata);
  };
}

var multiInject = /*#__PURE__*/injectBase(MULTI_INJECT_TAG);

function targetName(name) {
  return function (target, targetKey, index) {
    var metadata = new Metadata(NAME_TAG, name);
    tagParameter(target, targetKey, index, metadata);
  };
}

function propertyEventDecorator(eventKey, errorMessage) {
  return function () {
    return function (target, propertyKey) {
      var metadata = new Metadata(eventKey, propertyKey);

      if (reflection.hasOwnMetadata(eventKey, target.constructor)) {
        throw new Error(errorMessage);
      }

      reflection.defineMetadata(eventKey, metadata, target.constructor);
    };
  };
}

var postConstruct = /*#__PURE__*/propertyEventDecorator(POST_CONSTRUCT, MULTIPLE_POST_CONSTRUCT_METHODS);

var QueryableString = /*#__PURE__*/function () {
  function QueryableString(str) {
    this.str = void 0;
    this.str = str;
  }

  var _proto = QueryableString.prototype;

  _proto.startsWith = function startsWith(searchString) {
    return this.str.indexOf(searchString) === 0;
  };

  _proto.endsWith = function endsWith(searchString) {
    var reverseString = '';
    var reverseSearchString = searchString.split('').reverse().join('');
    reverseString = this.str.split('').reverse().join('');
    return this.startsWith.call({
      str: reverseString
    }, reverseSearchString);
  };

  _proto.contains = function contains(searchString) {
    return this.str.indexOf(searchString) !== -1;
  };

  _proto.equals = function equals(compareString) {
    return this.str === compareString;
  };

  _proto.value = function value() {
    return this.str;
  };

  return QueryableString;
}();

var Target = /*#__PURE__*/function () {
  function Target(type, identifier, serviceIdentifier, namedOrTagged) {
    this.id = void 0;
    this.type = void 0;
    this.serviceIdentifier = void 0;
    this.name = void 0;
    this.identifier = void 0;
    this.metadata = void 0;
    this.id = id();
    this.type = type;
    this.serviceIdentifier = serviceIdentifier;
    var queryableName = typeof identifier === 'symbol' ? getSymbolDescription(identifier) : identifier;
    this.name = new QueryableString(queryableName != null ? queryableName : '');
    this.identifier = identifier;
    this.metadata = new Array();
    var metadataItem = null; // is named target

    if (typeof namedOrTagged === 'string') {
      metadataItem = new Metadata(NAMED_TAG, namedOrTagged);
    } else if (namedOrTagged instanceof Metadata) {
      // is target with metadata
      metadataItem = namedOrTagged;
    } // target has metadata


    if (metadataItem !== null) {
      this.metadata.push(metadataItem);
    }
  }

  var _proto = Target.prototype;

  _proto.hasTag = function hasTag(key) {
    for (var _iterator = _createForOfIteratorHelperLoose(this.metadata), _step; !(_step = _iterator()).done;) {
      var m = _step.value;

      if (m.key === key) {
        return true;
      }
    }

    return false;
  };

  _proto.isArray = function isArray() {
    return this.hasTag(MULTI_INJECT_TAG);
  };

  _proto.matchesArray = function matchesArray(name) {
    return this.matchesTag(MULTI_INJECT_TAG)(name);
  };

  _proto.isNamed = function isNamed() {
    return this.hasTag(NAMED_TAG);
  };

  _proto.isTagged = function isTagged() {
    return this.metadata.some(function (metadata) {
      return NON_CUSTOM_TAG_KEYS.every(function (key) {
        return metadata.key !== key;
      });
    });
  };

  _proto.isOptional = function isOptional() {
    return this.matchesTag(OPTIONAL_TAG)(true);
  };

  _proto.getNamedTag = function getNamedTag() {
    if (this.isNamed()) {
      return this.metadata.filter(function (m) {
        return m.key === NAMED_TAG;
      })[0];
    }

    return null;
  };

  _proto.getCustomTags = function getCustomTags() {
    if (this.isTagged()) {
      return this.metadata.filter(function (metadata) {
        return NON_CUSTOM_TAG_KEYS.every(function (key) {
          return metadata.key !== key;
        });
      });
    }

    return null;
  };

  _proto.matchesNamedTag = function matchesNamedTag(name) {
    return this.matchesTag(NAMED_TAG)(name);
  };

  _proto.matchesTag = function matchesTag(key) {
    var _this = this;

    return function (value) {
      for (var _iterator2 = _createForOfIteratorHelperLoose(_this.metadata), _step2; !(_step2 = _iterator2()).done;) {
        var m = _step2.value;

        if (m.key === key && m.value === value) {
          return true;
        }
      }

      return false;
    };
  };

  return Target;
}();

function getDependencies(metadataReader, func) {
  var constructorName = getFunctionName(func);
  return getTargets(metadataReader, constructorName, func, false);
}

function getTargets(metadataReader, constructorName, func, isBaseClass) {
  var metadata = metadataReader.getConstructorMetadata(func); // TypeScript compiler generated annotations

  var serviceIdentifiers = metadata.compilerGeneratedMetadata; // All types resolved must be annotated with @injectable

  if (serviceIdentifiers === undefined) {
    var msg = MISSING_INJECTABLE_ANNOTATION + " " + constructorName + ".";
    throw new Error(msg);
  } // User generated annotations


  var constructorArgsMetadata = metadata.userGeneratedMetadata;
  var keys = Object.keys(constructorArgsMetadata);
  var hasUserDeclaredUnknownInjections = func.length === 0 && keys.length > 0;
  var hasOptionalParameters = keys.length > func.length;
  var iterations = hasUserDeclaredUnknownInjections || hasOptionalParameters ? keys.length : func.length; // Target instances that represent constructor arguments to be injected

  var constructorTargets = getConstructorArgsAsTargets(isBaseClass, constructorName, serviceIdentifiers, constructorArgsMetadata, iterations); // Target instances that represent properties to be injected

  var propertyTargets = getClassPropsAsTargets(metadataReader, func, constructorName);
  var targets = [].concat(constructorTargets, propertyTargets);
  return targets;
}

function getConstructorArgsAsTarget(index, isBaseClass, constructorName, serviceIdentifiers, constructorArgsMetadata) {
  var _constructorArgsMetad, _metadata$inject;

  // Create map from array of metadata for faster access to metadata
  var targetMetadata = (_constructorArgsMetad = constructorArgsMetadata[index.toString()]) != null ? _constructorArgsMetad : [];
  var metadata = formatTargetMetadata(targetMetadata);
  var isManaged = metadata.unmanaged !== true; // Take types to be injected from user-generated metadata
  // if not available use compiler-generated metadata

  var serviceIdentifier = serviceIdentifiers[index];
  var injectIdentifier = (_metadata$inject = metadata.inject) != null ? _metadata$inject : metadata.multiInject;
  serviceIdentifier = injectIdentifier ? injectIdentifier : serviceIdentifier; // we unwrap LazyServiceIdentifer wrappers to allow circular dependencies on symbols

  if (serviceIdentifier instanceof LazyServiceIdentifer) {
    // @ts-ignore
    serviceIdentifier = serviceIdentifier.unwrap();
  } // Types Object and Function are too ambiguous to be resolved
  // user needs to generate metadata manually for those


  if (isManaged) {
    var isObject = serviceIdentifier === Object;
    var isFunction = serviceIdentifier === Function;
    var isUndefined = serviceIdentifier === undefined;
    var isUnknownType = isObject || isFunction || isUndefined;

    if (!isBaseClass && isUnknownType) {
      var msg = MISSING_INJECT_ANNOTATION + " argument " + index + " in class " + constructorName + ".";
      throw new Error(msg);
    }

    var target = new Target(TargetTypeEnum.ConstructorArgument, metadata.targetName, serviceIdentifier);
    target.metadata = targetMetadata;
    return target;
  }

  return null;
}

function getConstructorArgsAsTargets(isBaseClass, constructorName, serviceIdentifiers, constructorArgsMetadata, iterations) {
  var targets = [];

  for (var i = 0; i < iterations; i++) {
    var index = i;
    var target = getConstructorArgsAsTarget(index, isBaseClass, constructorName, serviceIdentifiers, constructorArgsMetadata);

    if (target !== null) {
      targets.push(target);
    }
  }

  return targets;
}

function _getServiceIdentifierForProperty(inject, multiInject, propertyName, className) {
  var serviceIdentifier = inject != null ? inject : multiInject;

  if (serviceIdentifier === undefined) {
    var msg = MISSING_INJECTABLE_ANNOTATION + " for property " + String(propertyName) + " in class " + className + ".";
    throw new Error(msg);
  }

  return serviceIdentifier;
}

function getClassPropsAsTargets(metadataReader, constructorFunc, constructorName) {
  var classPropsMetadata = metadataReader.getPropertiesMetadata(constructorFunc);
  var targets = [];
  var symbolKeys = Object.getOwnPropertySymbols(classPropsMetadata);
  var stringKeys = Object.keys(classPropsMetadata);
  var keys = stringKeys.concat(symbolKeys);

  for (var _iterator = _createForOfIteratorHelperLoose(keys), _step; !(_step = _iterator()).done;) {
    var _metadata$targetName;

    var key = _step.value;
    // the metadata for the property being injected
    var targetMetadata = classPropsMetadata[key]; // the metadata formatted for easier access

    var metadata = formatTargetMetadata(targetMetadata); // the name of the property being injected

    var identifier = (_metadata$targetName = metadata.targetName) != null ? _metadata$targetName : key; // Take types to be injected from user-generated metadata

    var serviceIdentifier = _getServiceIdentifierForProperty(metadata.inject, metadata.multiInject, key, constructorName); // The property target


    var target = new Target(TargetTypeEnum.ClassProperty, identifier, serviceIdentifier);
    target.metadata = targetMetadata;
    targets.push(target);
  } // Check if base class has injected properties


  var baseConstructor = Object.getPrototypeOf(constructorFunc.prototype).constructor;

  if (baseConstructor !== Object) {
    var baseTargets = getClassPropsAsTargets(metadataReader, baseConstructor, constructorName);
    targets = [].concat(targets, baseTargets);
  }

  return targets;
}

function getBaseClassDependencyCount(metadataReader, func) {
  var baseConstructor = Object.getPrototypeOf(func.prototype).constructor;

  if (baseConstructor !== Object) {
    // get targets for base class
    var baseConstructorName = getFunctionName(baseConstructor);
    var targets = getTargets(metadataReader, baseConstructorName, baseConstructor, true); // get unmanaged metadata

    var metadata = targets.map(function (t) {
      return t.metadata.filter(function (m) {
        return m.key === UNMANAGED_TAG;
      });
    }); // Compare the number of constructor arguments with the number of
    // unmanaged dependencies unmanaged dependencies are not required

    var unmanagedCount = [].concat.apply([], metadata).length;
    var dependencyCount = targets.length - unmanagedCount;

    if (dependencyCount > 0) {
      return dependencyCount;
    } else {
      return getBaseClassDependencyCount(metadataReader, baseConstructor);
    }
  } else {
    return 0;
  }
}

function formatTargetMetadata(targetMetadata) {
  // Create map from array of metadata for faster access to metadata
  var targetMetadataMap = {};
  targetMetadata.forEach(function (m) {
    targetMetadataMap[m.key.toString()] = m.value;
  }); // user generated metadata

  return {
    inject: targetMetadataMap[INJECT_TAG],
    multiInject: targetMetadataMap[MULTI_INJECT_TAG],
    targetName: targetMetadataMap[NAME_TAG],
    unmanaged: targetMetadataMap[UNMANAGED_TAG]
  };
}

var Request = /*#__PURE__*/function () {
  function Request(serviceIdentifier, parentContext, parentRequest, bindings, target) {
    this.id = void 0;
    this.serviceIdentifier = void 0;
    this.parentContext = void 0;
    this.parentRequest = void 0;
    this.bindings = void 0;
    this.childRequests = void 0;
    this.target = void 0;
    this.requestScope = void 0;
    this.id = id();
    this.serviceIdentifier = serviceIdentifier;
    this.parentContext = parentContext;
    this.parentRequest = parentRequest;
    this.target = target;
    this.childRequests = [];
    this.bindings = Array.isArray(bindings) ? bindings : [bindings]; // Set requestScope if Request is the root request

    this.requestScope = parentRequest === null ? new Map() : null;
  }

  var _proto = Request.prototype;

  _proto.addChildRequest = function addChildRequest(serviceIdentifier, bindings, target) {
    var child = new Request(serviceIdentifier, this.parentContext, this, bindings, target);
    this.childRequests.push(child);
    return child;
  };

  return Request;
}();

function getBindingDictionary(container) {
  return container._bindingDictionary;
}

function _createTarget(isMultiInject, targetType, serviceIdentifier, name, key, value) {
  var metadataKey = isMultiInject ? MULTI_INJECT_TAG : INJECT_TAG;
  var injectMetadata = new Metadata(metadataKey, serviceIdentifier);
  var target = new Target(targetType, name, serviceIdentifier, injectMetadata);

  if (key !== undefined) {
    var tagMetadata = new Metadata(key, value);
    target.metadata.push(tagMetadata);
  }

  return target;
}

function _getActiveBindings(metadataReader, avoidConstraints, context, parentRequest, target) {
  var bindings = getBindings(context.container, target.serviceIdentifier);
  var activeBindings = []; // automatic binding

  if (bindings.length === BindingCount.NoBindingsAvailable && context.container.options.autoBindInjectable && typeof target.serviceIdentifier === 'function' && metadataReader.getConstructorMetadata(target.serviceIdentifier).compilerGeneratedMetadata) {
    context.container.bind(target.serviceIdentifier).toSelf();
    bindings = getBindings(context.container, target.serviceIdentifier);
  } // multiple bindings available


  if (!avoidConstraints) {
    // apply constraints if available to reduce the number of active bindings
    activeBindings = bindings.filter(function (binding) {
      var request = new Request(binding.serviceIdentifier, context, parentRequest, binding, target);
      return binding.constraint(request);
    });
  } else {
    // simple injection or multi-injection without constraints
    activeBindings = bindings;
  } // validate active bindings


  _validateActiveBindingCount(target.serviceIdentifier, activeBindings, target, context.container);

  return activeBindings;
}

function _validateActiveBindingCount(serviceIdentifier, bindings, target, container) {
  switch (bindings.length) {
    case BindingCount.NoBindingsAvailable:
      if (target.isOptional()) {
        return bindings;
      } else {
        var serviceIdentifierString = getServiceIdentifierAsString(serviceIdentifier);
        var msg = NOT_REGISTERED;
        msg += listMetadataForTarget(serviceIdentifierString, target);
        msg += listRegisteredBindingsForServiceIdentifier(container, serviceIdentifierString, getBindings);
        throw new Error(msg);
      }

    // @ts-ignore

    case BindingCount.OnlyOneBindingAvailable:
      return bindings;
    // eslint-disable no-fallthrough

    case BindingCount.MultipleBindingsAvailable:
    default:
      if (!target.isArray()) {
        var _serviceIdentifierString = getServiceIdentifierAsString(serviceIdentifier);

        var _msg = AMBIGUOUS_MATCH + " " + _serviceIdentifierString;

        _msg += listRegisteredBindingsForServiceIdentifier(container, _serviceIdentifierString, getBindings);
        throw new Error(_msg);
      } else {
        return bindings;
      }

  }
}

function _createSubRequests(metadataReader, avoidConstraints, serviceIdentifier, context, parentRequest, target) {
  var activeBindings;
  var childRequest;

  if (parentRequest === null) {
    activeBindings = _getActiveBindings(metadataReader, avoidConstraints, context, null, target);
    childRequest = new Request(serviceIdentifier, context, null, activeBindings, target);
    var thePlan = new Plan(context, childRequest);
    context.addPlan(thePlan);
  } else {
    activeBindings = _getActiveBindings(metadataReader, avoidConstraints, context, parentRequest, target); // @ts-ignore

    childRequest = parentRequest.addChildRequest(target.serviceIdentifier, activeBindings, target);
  }

  activeBindings.forEach(function (binding) {
    var subChildRequest = null;

    if (target.isArray()) {
      // @ts-ignore
      subChildRequest = childRequest.addChildRequest(binding.serviceIdentifier, binding, target);
    } else {
      if (binding.cache) {
        return;
      }

      subChildRequest = childRequest;
    }

    if (binding.type === BindingTypeEnum.Instance && binding.implementationType !== null) {
      var dependencies = getDependencies(metadataReader, binding.implementationType);

      if (!context.container.options.skipBaseClassChecks) {
        // Throw if a derived class does not implement its constructor explicitly
        // We do this to prevent errors when a base class (parent) has dependencies
        // and one of the derived classes (children) has no dependencies
        var baseClassDependencyCount = getBaseClassDependencyCount(metadataReader, binding.implementationType);

        if (dependencies.length < baseClassDependencyCount) {
          var error = ARGUMENTS_LENGTH_MISMATCH(getFunctionName(binding.implementationType));
          throw new Error(error);
        }
      }

      dependencies.forEach(function (dependency) {
        _createSubRequests(metadataReader, false, dependency.serviceIdentifier, context, subChildRequest, dependency);
      });
    }
  });
}

function getBindings(container, serviceIdentifier) {
  var bindings = [];
  var bindingDictionary = getBindingDictionary(container);

  if (bindingDictionary.hasKey(serviceIdentifier)) {
    bindings = bindingDictionary.get(serviceIdentifier);
  } else if (container.parent !== null) {
    // recursively try to get bindings from parent container
    bindings = getBindings(container.parent, serviceIdentifier);
  }

  return bindings;
}

function plan(metadataReader, container, isMultiInject, targetType, serviceIdentifier, key, value, avoidConstraints) {
  if (avoidConstraints === void 0) {
    avoidConstraints = false;
  }

  var context = new Context(container);

  var target = _createTarget(isMultiInject, targetType, serviceIdentifier, '', key, value);

  try {
    _createSubRequests(metadataReader, avoidConstraints, serviceIdentifier, context, null, target);

    return context;
  } catch (error) {
    if (error instanceof Error && isStackOverflowExeption(error)) {
      if (context.plan) {
        circularDependencyToException(context.plan.rootRequest);
      }
    }

    throw error;
  }
}
function createMockRequest(container, serviceIdentifier, key, value) {
  var target = new Target(TargetTypeEnum.Variable, '', serviceIdentifier, new Metadata(key, value));
  var context = new Context(container);
  var request = new Request(serviceIdentifier, context, null, [], target);
  return request;
}

var tryGetFromScope = function tryGetFromScope(requestScope, binding) {
  if (binding.scope === BindingScopeEnum.Singleton && binding.activated) {
    return binding.cache;
  }

  if (binding.scope === BindingScopeEnum.Request && requestScope.has(binding.id)) {
    return requestScope.get(binding.id);
  }

  return null;
};
var saveToScope = function saveToScope(requestScope, binding, result) {
  if (binding.scope === BindingScopeEnum.Singleton) {
    _saveToSingletonScope(binding, result);
  }

  if (binding.scope === BindingScopeEnum.Request) {
    _saveToRequestScope(requestScope, binding, result);
  }
};

var _saveToRequestScope = function _saveToRequestScope(requestScope, binding, result) {
  if (!requestScope.has(binding.id)) {
    requestScope.set(binding.id, result);
  }
};

var _saveToSingletonScope = function _saveToSingletonScope(binding, result) {
  // store in cache if scope is singleton
  binding.cache = result;
  binding.activated = true;

  if (isPromise(result)) {
    void _saveAsyncResultToSingletonScope(binding, result);
  }
};

var _saveAsyncResultToSingletonScope = /*#__PURE__*/function () {
  var _ref = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee(binding, asyncResult) {
    var result;
    return _regeneratorRuntime().wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return asyncResult;

          case 3:
            result = _context.sent;
            binding.cache = result;
            _context.next = 12;
            break;

          case 7:
            _context.prev = 7;
            _context.t0 = _context["catch"](0);
            // allow binding to retry in future
            binding.cache = null;
            binding.activated = false;
            throw _context.t0;

          case 12:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, null, [[0, 7]]);
  }));

  return function _saveAsyncResultToSingletonScope(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();

function _resolveRequests(childRequests, resolveRequest) {
  return childRequests.reduce(function (resolvedRequests, childRequest) {
    var injection = resolveRequest(childRequest);
    var targetType = childRequest.target.type;

    if (targetType === TargetTypeEnum.ConstructorArgument) {
      resolvedRequests.constructorInjections.push(injection);
    } else {
      resolvedRequests.propertyRequests.push(childRequest);
      resolvedRequests.propertyInjections.push(injection);
    }

    if (!resolvedRequests.isAsync) {
      resolvedRequests.isAsync = isPromiseOrContainsPromise(injection);
    }

    return resolvedRequests;
  }, {
    constructorInjections: [],
    propertyInjections: [],
    propertyRequests: [],
    isAsync: false
  });
}

function _createInstance(constr, childRequests, resolveRequest) {
  var result;

  if (childRequests.length > 0) {
    var resolved = _resolveRequests(childRequests, resolveRequest);

    var createInstanceWithInjectionsArg = _extends({}, resolved, {
      constr: constr
    });

    if (resolved.isAsync) {
      result = createInstanceWithInjectionsAsync(createInstanceWithInjectionsArg);
    } else {
      result = createInstanceWithInjections(createInstanceWithInjectionsArg);
    }
  } else {
    result = new constr();
  }

  return result;
}

function createInstanceWithInjections(args) {
  // @ts-ignore
  var instance = _construct(args.constr, args.constructorInjections);

  args.propertyRequests.forEach(function (r, index) {
    var property = r.target.identifier;
    var injection = args.propertyInjections[index];
    instance[property] = injection;
  });
  return instance;
}

function createInstanceWithInjectionsAsync(_x) {
  return _createInstanceWithInjectionsAsync.apply(this, arguments);
}

function _createInstanceWithInjectionsAsync() {
  _createInstanceWithInjectionsAsync = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee(args) {
    var constructorInjections, propertyInjections;
    return _regeneratorRuntime().wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return possiblyWaitInjections(args.constructorInjections);

          case 2:
            constructorInjections = _context.sent;
            _context.next = 5;
            return possiblyWaitInjections(args.propertyInjections);

          case 5:
            propertyInjections = _context.sent;
            return _context.abrupt("return", createInstanceWithInjections(_extends({}, args, {
              constructorInjections: constructorInjections,
              propertyInjections: propertyInjections
            })));

          case 7:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _createInstanceWithInjectionsAsync.apply(this, arguments);
}

function possiblyWaitInjections(_x2) {
  return _possiblyWaitInjections.apply(this, arguments);
}

function _possiblyWaitInjections() {
  _possiblyWaitInjections = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee2(possiblePromiseinjections) {
    var injections, _iterator, _step, injection;

    return _regeneratorRuntime().wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            injections = [];

            for (_iterator = _createForOfIteratorHelperLoose(possiblePromiseinjections); !(_step = _iterator()).done;) {
              injection = _step.value;

              if (Array.isArray(injection)) {
                injections.push(Promise.all(injection));
              } else {
                injections.push(injection);
              }
            }

            return _context2.abrupt("return", Promise.all(injections));

          case 3:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2);
  }));
  return _possiblyWaitInjections.apply(this, arguments);
}

function _getInstanceAfterPostConstruct(constr, result) {
  var postConstructResult = _postConstruct(constr, result);

  if (isPromise(postConstructResult)) {
    return postConstructResult.then(function () {
      return result;
    });
  } else {
    return result;
  }
}

function _postConstruct(constr, instance) {
  if (reflection.hasMetadata(POST_CONSTRUCT, constr)) {
    // @ts-ignore
    var data = reflection.getMetadata(POST_CONSTRUCT, constr);

    try {
      var _instance$data$value;

      return (_instance$data$value = instance[data.value]) == null ? void 0 : _instance$data$value.call(instance);
    } catch (e) {
      throw new Error(POST_CONSTRUCT_ERROR(constr.name, e.message));
    }
  }
}

function _validateInstanceResolution(binding, constr) {
  if (binding.scope !== BindingScopeEnum.Singleton) {
    _throwIfHandlingDeactivation(binding, constr);
  }
}

function _throwIfHandlingDeactivation(binding, constr) {
  var scopeErrorMessage = "Class cannot be instantiated in " + (binding.scope === BindingScopeEnum.Request ? 'request' : 'transient') + " scope.";

  if (typeof binding.onDeactivation === 'function') {
    throw new Error(ON_DEACTIVATION_ERROR(constr.name, scopeErrorMessage));
  }

  if (reflection.hasMetadata(PRE_DESTROY, constr)) {
    throw new Error(PRE_DESTROY_ERROR(constr.name, scopeErrorMessage));
  }
}

function resolveInstance(binding, constr, childRequests, resolveRequest) {
  _validateInstanceResolution(binding, constr);

  var result = _createInstance(constr, childRequests, resolveRequest);

  if (isPromise(result)) {
    return result.then(function (resolvedResult) {
      return _getInstanceAfterPostConstruct(constr, resolvedResult);
    });
  } else {
    return _getInstanceAfterPostConstruct(constr, result);
  }
}

var _resolveRequest = function _resolveRequest(requestScope) {
  return function (request) {
    request.parentContext.setCurrentRequest(request);
    var bindings = request.bindings;
    var childRequests = request.childRequests;
    var targetIsAnArray = request.target && request.target.isArray();
    var targetParentIsNotAnArray = !request.parentRequest || !request.parentRequest.target || !request.target || !request.parentRequest.target.matchesArray(request.target.serviceIdentifier);

    if (targetIsAnArray && targetParentIsNotAnArray) {
      // Create an array instead of creating an instance
      return childRequests.map(function (childRequest) {
        var _f = _resolveRequest(requestScope);

        return _f(childRequest);
      });
    } else {
      if (request.target.isOptional() && bindings.length === 0) {
        return undefined;
      }

      var binding = bindings[0];
      return _resolveBinding(requestScope, request, binding);
    }
  };
};

var _resolveFactoryFromBinding = function _resolveFactoryFromBinding(binding, context) {
  var factoryDetails = getFactoryDetails(binding);
  return tryAndThrowErrorIfStackOverflow(function () {
    return factoryDetails.factory.bind(binding)(context);
  }, function () {
    return new Error(CIRCULAR_DEPENDENCY_IN_FACTORY(factoryDetails.factoryType, context.currentRequest.serviceIdentifier.toString()));
  });
};

var _getResolvedFromBinding = function _getResolvedFromBinding(requestScope, request, binding) {
  var result;
  var childRequests = request.childRequests;
  ensureFullyBound(binding);

  switch (binding.type) {
    case BindingTypeEnum.ConstantValue:
    case BindingTypeEnum.Function:
      result = binding.cache;
      break;

    case BindingTypeEnum.Constructor:
      result = binding.implementationType;
      break;

    case BindingTypeEnum.Instance:
      result = resolveInstance(binding, binding.implementationType, childRequests, _resolveRequest(requestScope));
      break;

    default:
      result = _resolveFactoryFromBinding(binding, request.parentContext);
  }

  return result;
};

var _resolveInScope = function _resolveInScope(requestScope, binding, resolveFromBinding) {
  var result = tryGetFromScope(requestScope, binding);

  if (result !== null) {
    return result;
  }

  result = resolveFromBinding();
  saveToScope(requestScope, binding, result);
  return result;
};

var _resolveBinding = function _resolveBinding(requestScope, request, binding) {
  return _resolveInScope(requestScope, binding, function () {
    var result = _getResolvedFromBinding(requestScope, request, binding);

    if (isPromise(result)) {
      result = result.then(function (resolved) {
        return _onActivation(request, binding, resolved);
      });
    } else {
      result = _onActivation(request, binding, result);
    }

    return result;
  });
};

function _onActivation(request, binding, resolved) {
  var result = _bindingActivation(request.parentContext, binding, resolved);

  var containersIterator = _getContainersIterator(request.parentContext.container);

  var container;
  var containersIteratorResult = containersIterator.next();

  do {
    container = containersIteratorResult.value;
    var context = request.parentContext;
    var serviceIdentifier = request.serviceIdentifier;

    var activationsIterator = _getContainerActivationsForService(container, serviceIdentifier);

    if (isPromise(result)) {
      result = _activateContainerAsync(activationsIterator, context, result);
    } else {
      result = _activateContainer(activationsIterator, context, result);
    }

    containersIteratorResult = containersIterator.next(); // make sure if we are currently on the container that owns the binding, not to keep looping down to child containers
  } while (containersIteratorResult.done !== true && !getBindingDictionary(container).hasKey(request.serviceIdentifier));

  return result;
}

var _bindingActivation = function _bindingActivation(context, binding, previousResult) {
  var result; // use activation handler if available

  if (typeof binding.onActivation === 'function') {
    result = binding.onActivation(context, previousResult);
  } else {
    result = previousResult;
  }

  return result;
};

var _activateContainer = function _activateContainer(activationsIterator, context, result) {
  var activation = activationsIterator.next();

  while (!activation.done) {
    result = activation.value(context, result);

    if (isPromise(result)) {
      return _activateContainerAsync(activationsIterator, context, result);
    }

    activation = activationsIterator.next();
  }

  return result;
};

var _activateContainerAsync = /*#__PURE__*/function () {
  var _ref = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee(activationsIterator, context, resultPromise) {
    var result, activation;
    return _regeneratorRuntime().wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return resultPromise;

          case 2:
            result = _context.sent;
            activation = activationsIterator.next();

          case 4:
            if (activation.done) {
              _context.next = 11;
              break;
            }

            _context.next = 7;
            return activation.value(context, result);

          case 7:
            result = _context.sent;
            activation = activationsIterator.next();
            _context.next = 4;
            break;

          case 11:
            return _context.abrupt("return", result);

          case 12:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));

  return function _activateContainerAsync(_x, _x2, _x3) {
    return _ref.apply(this, arguments);
  };
}();

var _getContainerActivationsForService = function _getContainerActivationsForService(container, serviceIdentifier) {
  // smell accessing _activations, but similar pattern is done in planner.getBindingDictionary()
  var activations = container._activations;
  return activations.hasKey(serviceIdentifier) ? activations.get(serviceIdentifier).values() : [].values();
};

var _getContainersIterator = function _getContainersIterator(container) {
  var containersStack = [container];
  var parent = container.parent;

  while (parent !== null) {
    containersStack.push(parent);
    parent = parent.parent;
  }

  var getNextContainer = function getNextContainer() {
    var nextContainer = containersStack.pop();

    if (nextContainer !== undefined) {
      return {
        done: false,
        value: nextContainer
      };
    } else {
      return {
        done: true,
        value: undefined
      };
    }
  };

  var containersIterator = {
    next: getNextContainer
  };
  return containersIterator;
};

function resolve(context) {
  var _f = _resolveRequest(context.plan.rootRequest.requestScope);

  return _f(context.plan.rootRequest);
}

var traverseAncerstors = function traverseAncerstors(request, constraint) {
  var _request$parentReques;

  var parent = (_request$parentReques = request == null ? void 0 : request.parentRequest) != null ? _request$parentReques : null;

  if (parent !== null) {
    return constraint(parent) ? true : traverseAncerstors(parent, constraint);
  }

  return false;
};
/**
 * This helpers use currying to help you to generate constraints
 *
 * @param key
 */

var taggedConstraint = function taggedConstraint(key) {
  return function (value) {
    var constraint = function constraint(request) {
      var _request$target$match, _request$target;

      return (_request$target$match = request == null ? void 0 : (_request$target = request.target) == null ? void 0 : _request$target.matchesTag(key)(value)) != null ? _request$target$match : false;
    };

    constraint.metaData = new Metadata(key, value);
    return constraint;
  };
};
var namedConstraint = /*#__PURE__*/taggedConstraint(NAMED_TAG);
var typeConstraint = function typeConstraint(type) {
  return function (request) {
    // Using index 0 because constraints are applied
    // to one binding at a time (see Planner class)
    var binding = null;

    if (request !== null) {
      binding = request == null ? void 0 : request.bindings[0];

      if (typeof type === 'string') {
        var _binding;

        var serviceIdentifier = (_binding = binding) == null ? void 0 : _binding.serviceIdentifier;
        return serviceIdentifier === type;
      } else {
        var _constructor = request == null ? void 0 : request.bindings[0].implementationType;

        return type === _constructor;
      }
    }

    return false;
  };
};

var BindingWhenSyntax = /*#__PURE__*/function () {
  function BindingWhenSyntax(binding) {
    this._binding = void 0;
    this._binding = binding;
  }

  var _proto = BindingWhenSyntax.prototype;

  _proto.when = function when(constraint) {
    this._binding.constraint = constraint;
    return new BindingOnSyntax(this._binding);
  };

  _proto.whenTargetNamed = function whenTargetNamed(name) {
    this._binding.constraint = namedConstraint(name);
    return new BindingOnSyntax(this._binding);
  };

  _proto.whenTargetIsDefault = function whenTargetIsDefault() {
    // @ts-ignore
    this._binding.constraint = function (request) {
      var targetIsDefault = (request == null ? void 0 : request.target) !== null && !(request != null && request.target.isNamed()) && !(request != null && request.target.isTagged());
      return targetIsDefault;
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenTargetTagged = function whenTargetTagged(tag, value) {
    this._binding.constraint = taggedConstraint(tag)(value);
    return new BindingOnSyntax(this._binding);
  };

  _proto.whenInjectedInto = function whenInjectedInto(parent) {
    this._binding.constraint = function (request) {
      return request !== null && typeConstraint(parent)(request == null ? void 0 : request.parentRequest);
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenParentNamed = function whenParentNamed(name) {
    this._binding.constraint = function (request) {
      return request !== null && namedConstraint(name)(request == null ? void 0 : request.parentRequest);
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenParentTagged = function whenParentTagged(tag, value) {
    this._binding.constraint = function (request) {
      return request !== null && taggedConstraint(tag)(value)(request == null ? void 0 : request.parentRequest);
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenAnyAncestorIs = function whenAnyAncestorIs(ancestor) {
    this._binding.constraint = function (request) {
      return request !== null && traverseAncerstors(request, typeConstraint(ancestor));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenNoAncestorIs = function whenNoAncestorIs(ancestor) {
    this._binding.constraint = function (request) {
      return request !== null && !traverseAncerstors(request, typeConstraint(ancestor));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenAnyAncestorNamed = function whenAnyAncestorNamed(name) {
    this._binding.constraint = function (request) {
      return request !== null && traverseAncerstors(request, namedConstraint(name));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenNoAncestorNamed = function whenNoAncestorNamed(name) {
    this._binding.constraint = function (request) {
      return request !== null && !traverseAncerstors(request, namedConstraint(name));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenAnyAncestorTagged = function whenAnyAncestorTagged(tag, value) {
    this._binding.constraint = function (request) {
      return request !== null && traverseAncerstors(request, taggedConstraint(tag)(value));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenNoAncestorTagged = function whenNoAncestorTagged(tag, value) {
    this._binding.constraint = function (request) {
      return request !== null && !traverseAncerstors(request, taggedConstraint(tag)(value));
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenAnyAncestorMatches = function whenAnyAncestorMatches(constraint) {
    this._binding.constraint = function (request) {
      return request !== null && traverseAncerstors(request, constraint);
    };

    return new BindingOnSyntax(this._binding);
  };

  _proto.whenNoAncestorMatches = function whenNoAncestorMatches(constraint) {
    this._binding.constraint = function (request) {
      return request !== null && !traverseAncerstors(request, constraint);
    };

    return new BindingOnSyntax(this._binding);
  };

  return BindingWhenSyntax;
}();

var BindingOnSyntax = /*#__PURE__*/function () {
  function BindingOnSyntax(binding) {
    this._binding = void 0;
    this._binding = binding;
  }

  var _proto = BindingOnSyntax.prototype;

  _proto.onActivation = function onActivation(handler) {
    this._binding.onActivation = handler;
    return new BindingWhenSyntax(this._binding);
  };

  _proto.onDeactivation = function onDeactivation(handler) {
    this._binding.onDeactivation = handler;
    return new BindingWhenSyntax(this._binding);
  };

  return BindingOnSyntax;
}();

var BindingWhenOnSyntax = /*#__PURE__*/function () {
  function BindingWhenOnSyntax(binding) {
    this._bindingWhenSyntax = void 0;
    this._bindingOnSyntax = void 0;
    this._binding = void 0;
    this._binding = binding;
    this._bindingWhenSyntax = new BindingWhenSyntax(this._binding);
    this._bindingOnSyntax = new BindingOnSyntax(this._binding);
  }

  var _proto = BindingWhenOnSyntax.prototype;

  _proto.when = function when(constraint) {
    return this._bindingWhenSyntax.when(constraint);
  };

  _proto.whenTargetNamed = function whenTargetNamed(name) {
    return this._bindingWhenSyntax.whenTargetNamed(name);
  };

  _proto.whenTargetIsDefault = function whenTargetIsDefault() {
    return this._bindingWhenSyntax.whenTargetIsDefault();
  };

  _proto.whenTargetTagged = function whenTargetTagged(tag, value) {
    return this._bindingWhenSyntax.whenTargetTagged(tag, value);
  };

  _proto.whenInjectedInto = function whenInjectedInto(parent) {
    return this._bindingWhenSyntax.whenInjectedInto(parent);
  };

  _proto.whenParentNamed = function whenParentNamed(name) {
    return this._bindingWhenSyntax.whenParentNamed(name);
  };

  _proto.whenParentTagged = function whenParentTagged(tag, value) {
    return this._bindingWhenSyntax.whenParentTagged(tag, value);
  };

  _proto.whenAnyAncestorIs = function whenAnyAncestorIs(ancestor) {
    return this._bindingWhenSyntax.whenAnyAncestorIs(ancestor);
  };

  _proto.whenNoAncestorIs = function whenNoAncestorIs(ancestor) {
    return this._bindingWhenSyntax.whenNoAncestorIs(ancestor);
  };

  _proto.whenAnyAncestorNamed = function whenAnyAncestorNamed(name) {
    return this._bindingWhenSyntax.whenAnyAncestorNamed(name);
  };

  _proto.whenAnyAncestorTagged = function whenAnyAncestorTagged(tag, value) {
    return this._bindingWhenSyntax.whenAnyAncestorTagged(tag, value);
  };

  _proto.whenNoAncestorNamed = function whenNoAncestorNamed(name) {
    return this._bindingWhenSyntax.whenNoAncestorNamed(name);
  };

  _proto.whenNoAncestorTagged = function whenNoAncestorTagged(tag, value) {
    return this._bindingWhenSyntax.whenNoAncestorTagged(tag, value);
  };

  _proto.whenAnyAncestorMatches = function whenAnyAncestorMatches(constraint) {
    return this._bindingWhenSyntax.whenAnyAncestorMatches(constraint);
  };

  _proto.whenNoAncestorMatches = function whenNoAncestorMatches(constraint) {
    return this._bindingWhenSyntax.whenNoAncestorMatches(constraint);
  };

  _proto.onActivation = function onActivation(handler) {
    return this._bindingOnSyntax.onActivation(handler);
  };

  _proto.onDeactivation = function onDeactivation(handler) {
    return this._bindingOnSyntax.onDeactivation(handler);
  };

  return BindingWhenOnSyntax;
}();

var BindingInSyntax = /*#__PURE__*/function () {
  function BindingInSyntax(binding) {
    this._binding = void 0;
    this._binding = binding;
  }

  var _proto = BindingInSyntax.prototype;

  _proto.inRequestScope = function inRequestScope() {
    this._binding.scope = BindingScopeEnum.Request;
    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.inSingletonScope = function inSingletonScope() {
    this._binding.scope = BindingScopeEnum.Singleton;
    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.inTransientScope = function inTransientScope() {
    this._binding.scope = BindingScopeEnum.Transient;
    return new BindingWhenOnSyntax(this._binding);
  };

  return BindingInSyntax;
}();

var BindingInWhenOnSyntax = /*#__PURE__*/function () {
  function BindingInWhenOnSyntax(binding) {
    this._bindingInSyntax = void 0;
    this._bindingWhenSyntax = void 0;
    this._bindingOnSyntax = void 0;
    this._binding = void 0;
    this._binding = binding;
    this._bindingWhenSyntax = new BindingWhenSyntax(this._binding);
    this._bindingOnSyntax = new BindingOnSyntax(this._binding);
    this._bindingInSyntax = new BindingInSyntax(binding);
  }

  var _proto = BindingInWhenOnSyntax.prototype;

  _proto.inRequestScope = function inRequestScope() {
    return this._bindingInSyntax.inRequestScope();
  };

  _proto.inSingletonScope = function inSingletonScope() {
    return this._bindingInSyntax.inSingletonScope();
  };

  _proto.inTransientScope = function inTransientScope() {
    return this._bindingInSyntax.inTransientScope();
  };

  _proto.when = function when(constraint) {
    return this._bindingWhenSyntax.when(constraint);
  };

  _proto.whenTargetNamed = function whenTargetNamed(name) {
    return this._bindingWhenSyntax.whenTargetNamed(name);
  };

  _proto.whenTargetIsDefault = function whenTargetIsDefault() {
    return this._bindingWhenSyntax.whenTargetIsDefault();
  };

  _proto.whenTargetTagged = function whenTargetTagged(tag, value) {
    return this._bindingWhenSyntax.whenTargetTagged(tag, value);
  };

  _proto.whenInjectedInto = function whenInjectedInto(parent) {
    return this._bindingWhenSyntax.whenInjectedInto(parent);
  };

  _proto.whenParentNamed = function whenParentNamed(name) {
    return this._bindingWhenSyntax.whenParentNamed(name);
  };

  _proto.whenParentTagged = function whenParentTagged(tag, value) {
    return this._bindingWhenSyntax.whenParentTagged(tag, value);
  };

  _proto.whenAnyAncestorIs = function whenAnyAncestorIs(ancestor) {
    return this._bindingWhenSyntax.whenAnyAncestorIs(ancestor);
  };

  _proto.whenNoAncestorIs = function whenNoAncestorIs(ancestor) {
    return this._bindingWhenSyntax.whenNoAncestorIs(ancestor);
  };

  _proto.whenAnyAncestorNamed = function whenAnyAncestorNamed(name) {
    return this._bindingWhenSyntax.whenAnyAncestorNamed(name);
  };

  _proto.whenAnyAncestorTagged = function whenAnyAncestorTagged(tag, value) {
    return this._bindingWhenSyntax.whenAnyAncestorTagged(tag, value);
  };

  _proto.whenNoAncestorNamed = function whenNoAncestorNamed(name) {
    return this._bindingWhenSyntax.whenNoAncestorNamed(name);
  };

  _proto.whenNoAncestorTagged = function whenNoAncestorTagged(tag, value) {
    return this._bindingWhenSyntax.whenNoAncestorTagged(tag, value);
  };

  _proto.whenAnyAncestorMatches = function whenAnyAncestorMatches(constraint) {
    return this._bindingWhenSyntax.whenAnyAncestorMatches(constraint);
  };

  _proto.whenNoAncestorMatches = function whenNoAncestorMatches(constraint) {
    return this._bindingWhenSyntax.whenNoAncestorMatches(constraint);
  };

  _proto.onActivation = function onActivation(handler) {
    return this._bindingOnSyntax.onActivation(handler);
  };

  _proto.onDeactivation = function onDeactivation(handler) {
    return this._bindingOnSyntax.onDeactivation(handler);
  };

  return BindingInWhenOnSyntax;
}();

var BindingToSyntax = /*#__PURE__*/function () {
  function BindingToSyntax(binding) {
    this._binding = void 0;
    this._binding = binding;
  } // @ts-ignore


  var _proto = BindingToSyntax.prototype;

  _proto.to = function to(constructor) {
    this._binding.type = BindingTypeEnum.Instance; // @ts-ignore

    this._binding.implementationType = constructor;
    return new BindingInWhenOnSyntax(this._binding);
  };

  _proto.toSelf = function toSelf() {
    if (typeof this._binding.serviceIdentifier !== 'function') {
      throw new Error("" + INVALID_TO_SELF_VALUE);
    }

    var self = this._binding.serviceIdentifier; // @ts-ignore

    return this.to(self);
  };

  _proto.toConstantValue = function toConstantValue(value) {
    this._binding.type = BindingTypeEnum.ConstantValue;
    this._binding.cache = value;
    this._binding.dynamicValue = null;
    this._binding.implementationType = null;
    this._binding.scope = BindingScopeEnum.Singleton; // @ts-ignore

    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.toDynamicValue = function toDynamicValue(func) {
    this._binding.type = BindingTypeEnum.DynamicValue;
    this._binding.cache = null;
    this._binding.dynamicValue = func;
    this._binding.implementationType = null; // @ts-ignore

    return new BindingInWhenOnSyntax(this._binding);
  };

  _proto.toConstructor = function toConstructor(constructor) {
    this._binding.type = BindingTypeEnum.Constructor;
    this._binding.implementationType = constructor;
    this._binding.scope = BindingScopeEnum.Singleton; // @ts-ignore

    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.toFactory = function toFactory(factory) {
    this._binding.type = BindingTypeEnum.Factory;
    this._binding.factory = factory;
    this._binding.scope = BindingScopeEnum.Singleton; // @ts-ignore

    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.toFunction = function toFunction(func) {
    // toFunction is an alias of toConstantValue
    if (typeof func !== 'function') {
      throw new Error(INVALID_FUNCTION_BINDING);
    }

    var bindingWhenOnSyntax = this.toConstantValue(func);
    this._binding.type = BindingTypeEnum.Function;
    this._binding.scope = BindingScopeEnum.Singleton;
    return bindingWhenOnSyntax;
  };

  _proto.toAutoFactory = function toAutoFactory(serviceIdentifier) {
    this._binding.type = BindingTypeEnum.Factory;

    this._binding.factory = function (context) {
      var autofactory = function autofactory() {
        return context.container.get(serviceIdentifier);
      };

      return autofactory;
    };

    this._binding.scope = BindingScopeEnum.Singleton; // @ts-ignore

    return new BindingWhenOnSyntax(this._binding);
  } // @ts-ignore
  ;

  _proto.toAutoNamedFactory = function toAutoNamedFactory(serviceIdentifier) {
    this._binding.type = BindingTypeEnum.Factory;

    this._binding.factory = function (context) {
      return function (named) {
        return context.container.getNamed(serviceIdentifier, named);
      };
    };

    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.toProvider = function toProvider(provider) {
    this._binding.type = BindingTypeEnum.Provider;
    this._binding.provider = provider; // @ts-ignore

    return new BindingWhenOnSyntax(this._binding);
  };

  _proto.toService = function toService(service) {
    this.toDynamicValue(function (context) {
      return context.container.get(service);
    });
  };

  return BindingToSyntax;
}();

var ContainerSnapshot = /*#__PURE__*/function () {
  function ContainerSnapshot() {
    this.bindings = void 0;
    this.activations = void 0;
    this.deactivations = void 0;
    this.middleware = void 0;
    this.moduleActivationStore = void 0;
  }

  ContainerSnapshot.of = function of(bindings, middleware, activations, deactivations, moduleActivationStore) {
    var snapshot = new ContainerSnapshot();
    snapshot.bindings = bindings;
    snapshot.middleware = middleware;
    snapshot.deactivations = deactivations;
    snapshot.activations = activations;
    snapshot.moduleActivationStore = moduleActivationStore;
    return snapshot;
  };

  return ContainerSnapshot;
}();

var Lookup = /*#__PURE__*/function () {
  // dictionary used store multiple values for each key <key>
  function Lookup() {
    this._map = void 0;
    this._map = new Map();
  }

  var _proto = Lookup.prototype;

  _proto.getMap = function getMap() {
    return this._map;
  } // adds a new entry to _map
  ;

  _proto.add = function add(serviceIdentifier, value) {
    if (serviceIdentifier === null || serviceIdentifier === undefined) {
      throw new Error(NULL_ARGUMENT);
    }

    if (value === null || value === undefined) {
      throw new Error(NULL_ARGUMENT);
    }

    var entry = this._map.get(serviceIdentifier);

    if (entry !== undefined) {
      entry.push(value);
    } else {
      this._map.set(serviceIdentifier, [value]);
    }
  } // gets the value of a entry by its key (serviceIdentifier)
  ;

  _proto.get = function get(serviceIdentifier) {
    if (serviceIdentifier === null || serviceIdentifier === undefined) {
      throw new Error(NULL_ARGUMENT);
    }

    var entry = this._map.get(serviceIdentifier);

    if (entry !== undefined) {
      return entry;
    } else {
      throw new Error(KEY_NOT_FOUND);
    }
  } // removes a entry from _map by its key (serviceIdentifier)
  ;

  _proto.remove = function remove(serviceIdentifier) {
    if (serviceIdentifier === null || serviceIdentifier === undefined) {
      throw new Error(NULL_ARGUMENT);
    }

    if (!this._map["delete"](serviceIdentifier)) {
      throw new Error(KEY_NOT_FOUND);
    }
  };

  _proto.removeIntersection = function removeIntersection(lookup) {
    var _this = this;

    this.traverse(function (serviceIdentifier, value) {
      var lookupActivations = lookup.hasKey(serviceIdentifier) ? lookup.get(serviceIdentifier) : undefined;

      if (lookupActivations !== undefined) {
        var filteredValues = value.filter(function (lookupValue) {
          return !lookupActivations.some(function (moduleActivation) {
            return lookupValue === moduleActivation;
          });
        });

        _this._setValue(serviceIdentifier, filteredValues);
      }
    });
  };

  _proto.removeByCondition = function removeByCondition(condition) {
    var _this2 = this;

    var removals = [];

    this._map.forEach(function (entries, key) {
      var updatedEntries = [];

      for (var _iterator = _createForOfIteratorHelperLoose(entries), _step; !(_step = _iterator()).done;) {
        var entry = _step.value;
        var remove = condition(entry);

        if (remove) {
          removals.push(entry);
        } else {
          updatedEntries.push(entry);
        }
      }

      _this2._setValue(key, updatedEntries);
    });

    return removals;
  } // returns true if _map contains a key (serviceIdentifier)
  ;

  _proto.hasKey = function hasKey(serviceIdentifier) {
    if (serviceIdentifier === null || serviceIdentifier === undefined) {
      throw new Error(NULL_ARGUMENT);
    }

    return this._map.has(serviceIdentifier);
  } // returns a new Lookup instance; note: this is not a deep clone, only Lookup related data structure (dictionary) is
  // cloned, content remains the same
  ;

  _proto.clone = function clone() {
    var copy = new Lookup();

    this._map.forEach(function (value, key) {
      value.forEach(function (b) {
        return copy.add(key, isClonable(b) ? b.clone() : b);
      });
    });

    return copy;
  };

  _proto.traverse = function traverse(func) {
    this._map.forEach(function (value, key) {
      func(key, value);
    });
  };

  _proto._setValue = function _setValue(serviceIdentifier, value) {
    if (value.length > 0) {
      this._map.set(serviceIdentifier, value);
    } else {
      this._map["delete"](serviceIdentifier);
    }
  };

  return Lookup;
}();

var ModuleActivationStore = /*#__PURE__*/function () {
  function ModuleActivationStore() {
    this._map = new Map();
  }

  var _proto = ModuleActivationStore.prototype;

  _proto.remove = function remove(moduleId) {
    if (this._map.has(moduleId)) {
      var handlers = this._map.get(moduleId);

      this._map["delete"](moduleId);

      return handlers;
    }

    return this._getEmptyHandlersStore();
  };

  _proto.addDeactivation = function addDeactivation(moduleId, serviceIdentifier, onDeactivation) {
    this._getModuleActivationHandlers(moduleId).onDeactivations.add(serviceIdentifier, onDeactivation);
  };

  _proto.addActivation = function addActivation(moduleId, serviceIdentifier, onActivation) {
    this._getModuleActivationHandlers(moduleId).onActivations.add(serviceIdentifier, onActivation);
  };

  _proto.clone = function clone() {
    var clone = new ModuleActivationStore();

    this._map.forEach(function (handlersStore, moduleId) {
      clone._map.set(moduleId, {
        onActivations: handlersStore.onActivations.clone(),
        onDeactivations: handlersStore.onDeactivations.clone()
      });
    });

    return clone;
  };

  _proto._getModuleActivationHandlers = function _getModuleActivationHandlers(moduleId) {
    var moduleActivationHandlers = this._map.get(moduleId);

    if (moduleActivationHandlers === undefined) {
      moduleActivationHandlers = this._getEmptyHandlersStore();

      this._map.set(moduleId, moduleActivationHandlers);
    }

    return moduleActivationHandlers;
  };

  _proto._getEmptyHandlersStore = function _getEmptyHandlersStore() {
    var handlersStore = {
      onActivations: new Lookup(),
      onDeactivations: new Lookup()
    };
    return handlersStore;
  };

  return ModuleActivationStore;
}();

var Container = /*#__PURE__*/function () {
  Container.merge = function merge(container1, container2) {
    var container = new Container();

    for (var _len = arguments.length, containers = new Array(_len > 2 ? _len - 2 : 0), _key2 = 2; _key2 < _len; _key2++) {
      containers[_key2 - 2] = arguments[_key2];
    }

    var targetContainers = [container1, container2].concat(containers).map(function (targetContainer) {
      return getBindingDictionary(targetContainer);
    }); // @ts-ignore

    var bindingDictionary = getBindingDictionary(container);

    function copyDictionary(origin, destination) {
      origin.traverse(function (_key, value) {
        value.forEach(function (binding) {
          destination.add(binding.serviceIdentifier, binding.clone());
        });
      });
    }

    targetContainers.forEach(function (targetBindingDictionary) {
      copyDictionary(targetBindingDictionary, bindingDictionary);
    });
    return container;
  };

  function Container(containerOptions) {
    this.id = void 0;
    this.parent = void 0;
    this.options = void 0;
    this._middleware = void 0;
    this._bindingDictionary = void 0;
    this._activations = void 0;
    this._deactivations = void 0;
    this._snapshots = void 0;
    this._metadataReader = void 0;
    this._moduleActivationStore = void 0;
    var options = containerOptions != null ? containerOptions : {};

    if (typeof options !== 'object') {
      throw new Error("" + CONTAINER_OPTIONS_MUST_BE_AN_OBJECT);
    }

    if (options.defaultScope === undefined) {
      options.defaultScope = BindingScopeEnum.Transient;
    } else if (options.defaultScope !== BindingScopeEnum.Singleton && options.defaultScope !== BindingScopeEnum.Transient && options.defaultScope !== BindingScopeEnum.Request) {
      throw new Error("" + CONTAINER_OPTIONS_INVALID_DEFAULT_SCOPE);
    }

    if (options.autoBindInjectable === undefined) {
      options.autoBindInjectable = false;
    } else if (typeof options.autoBindInjectable !== 'boolean') {
      throw new Error("" + CONTAINER_OPTIONS_INVALID_AUTO_BIND_INJECTABLE);
    }

    if (options.skipBaseClassChecks === undefined) {
      options.skipBaseClassChecks = false;
    } else if (typeof options.skipBaseClassChecks !== 'boolean') {
      throw new Error("" + CONTAINER_OPTIONS_INVALID_SKIP_BASE_CHECK);
    }

    this.options = {
      autoBindInjectable: options.autoBindInjectable,
      defaultScope: options.defaultScope,
      skipBaseClassChecks: options.skipBaseClassChecks
    };
    this.id = id();
    this._bindingDictionary = new Lookup();
    this._snapshots = [];
    this._middleware = null;
    this._activations = new Lookup();
    this._deactivations = new Lookup();
    this.parent = null;
    this._metadataReader = new MetadataReader();
    this._moduleActivationStore = new ModuleActivationStore();
  }

  var _proto = Container.prototype;

  _proto.load = function load() {
    var getHelpers = this._getContainerModuleHelpersFactory();

    for (var _len2 = arguments.length, modules = new Array(_len2), _key3 = 0; _key3 < _len2; _key3++) {
      modules[_key3] = arguments[_key3];
    }

    for (var _i = 0, _modules = modules; _i < _modules.length; _i++) {
      var currentModule = _modules[_i];
      var containerModuleHelpers = getHelpers(currentModule.id);
      currentModule.registry(containerModuleHelpers.bindFunction, containerModuleHelpers.unbindFunction, containerModuleHelpers.isboundFunction, containerModuleHelpers.rebindFunction, containerModuleHelpers.unbindAsyncFunction, containerModuleHelpers.onActivationFunction, containerModuleHelpers.onDeactivationFunction);
    }
  };

  _proto.loadAsync = /*#__PURE__*/function () {
    var _loadAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee() {
      var getHelpers,
          _len3,
          modules,
          _key4,
          _i2,
          _modules2,
          currentModule,
          containerModuleHelpers,
          _args = arguments;

      return _regeneratorRuntime().wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              getHelpers = this._getContainerModuleHelpersFactory();

              for (_len3 = _args.length, modules = /*#__PURE__*/new Array(_len3), _key4 = 0; _key4 < _len3; _key4++) {
                modules[_key4] = _args[_key4];
              }

              _i2 = 0, _modules2 = modules;

            case 3:
              if (!(_i2 < _modules2.length)) {
                _context.next = 11;
                break;
              }

              currentModule = _modules2[_i2];
              containerModuleHelpers = getHelpers(currentModule.id);
              _context.next = 8;
              return currentModule.registry(containerModuleHelpers.bindFunction, containerModuleHelpers.unbindFunction, containerModuleHelpers.isboundFunction, containerModuleHelpers.rebindFunction, containerModuleHelpers.unbindAsyncFunction, containerModuleHelpers.onActivationFunction, containerModuleHelpers.onDeactivationFunction);

            case 8:
              _i2++;
              _context.next = 3;
              break;

            case 11:
            case "end":
              return _context.stop();
          }
        }
      }, _callee, this);
    }));

    function loadAsync() {
      return _loadAsync.apply(this, arguments);
    }

    return loadAsync;
  }();

  _proto.unload = function unload() {
    var _this = this;

    var conditionFactory = function conditionFactory(expected) {
      return function (item) {
        return item.moduleId === expected;
      };
    };

    for (var _len4 = arguments.length, modules = new Array(_len4), _key5 = 0; _key5 < _len4; _key5++) {
      modules[_key5] = arguments[_key5];
    }

    modules.forEach(function (module) {
      var condition = conditionFactory(module.id);

      _this._bindingDictionary.removeByCondition(condition);
    });
  } // Registers a type binding
  ;

  _proto.bind = function bind(serviceIdentifier) {
    var _this$options$default;

    var scope = (_this$options$default = this.options.defaultScope) != null ? _this$options$default : BindingScopeEnum.Transient;
    var binding = new Binding(serviceIdentifier, scope);

    this._bindingDictionary.add(serviceIdentifier, binding); // @ts-ignore


    return new BindingToSyntax(binding);
  };

  _proto.rebind = function rebind(serviceIdentifier) {
    this.unbind(serviceIdentifier);
    return this.bind(serviceIdentifier);
  };

  _proto.rebindAsync = /*#__PURE__*/function () {
    var _rebindAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee2(serviceIdentifier) {
      return _regeneratorRuntime().wrap(function _callee2$(_context2) {
        while (1) {
          switch (_context2.prev = _context2.next) {
            case 0:
              _context2.next = 2;
              return this.unbindAsync(serviceIdentifier);

            case 2:
              return _context2.abrupt("return", this.bind(serviceIdentifier));

            case 3:
            case "end":
              return _context2.stop();
          }
        }
      }, _callee2, this);
    }));

    function rebindAsync(_x) {
      return _rebindAsync.apply(this, arguments);
    }

    return rebindAsync;
  }() // Removes a type binding from the registry by its key
  ;

  _proto.unbind = function unbind(serviceIdentifier) {
    if (this._bindingDictionary.hasKey(serviceIdentifier)) {
      var bindings = this._bindingDictionary.get(serviceIdentifier);

      this._deactivateSingletons(bindings);
    }

    this._removeServiceFromDictionary(serviceIdentifier);
  };

  _proto.unbindAsync = /*#__PURE__*/function () {
    var _unbindAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee3(serviceIdentifier) {
      var bindings;
      return _regeneratorRuntime().wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              if (!this._bindingDictionary.hasKey(serviceIdentifier)) {
                _context3.next = 4;
                break;
              }

              bindings = this._bindingDictionary.get(serviceIdentifier);
              _context3.next = 4;
              return this._deactivateSingletonsAsync(bindings);

            case 4:
              this._removeServiceFromDictionary(serviceIdentifier);

            case 5:
            case "end":
              return _context3.stop();
          }
        }
      }, _callee3, this);
    }));

    function unbindAsync(_x2) {
      return _unbindAsync.apply(this, arguments);
    }

    return unbindAsync;
  }() // Removes all the type bindings from the registry
  ;

  _proto.unbindAll = function unbindAll() {
    var _this2 = this;

    this._bindingDictionary.traverse(function (_key, value) {
      _this2._deactivateSingletons(value);
    });

    this._bindingDictionary = new Lookup();
  };

  _proto.unbindAllAsync = /*#__PURE__*/function () {
    var _unbindAllAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee4() {
      var _this3 = this;

      var promises;
      return _regeneratorRuntime().wrap(function _callee4$(_context4) {
        while (1) {
          switch (_context4.prev = _context4.next) {
            case 0:
              promises = [];

              this._bindingDictionary.traverse(function (_key, value) {
                promises.push(_this3._deactivateSingletonsAsync(value));
              });

              _context4.next = 4;
              return Promise.all(promises);

            case 4:
              this._bindingDictionary = new Lookup();

            case 5:
            case "end":
              return _context4.stop();
          }
        }
      }, _callee4, this);
    }));

    function unbindAllAsync() {
      return _unbindAllAsync.apply(this, arguments);
    }

    return unbindAllAsync;
  }();

  _proto.onActivation = function onActivation(serviceIdentifier, _onActivation) {
    this._activations.add(serviceIdentifier, _onActivation);
  };

  _proto.onDeactivation = function onDeactivation(serviceIdentifier, _onDeactivation) {
    this._deactivations.add(serviceIdentifier, _onDeactivation);
  } // Allows to check if there are bindings available for serviceIdentifier
  ;

  _proto.isBound = function isBound(serviceIdentifier) {
    var bound = this._bindingDictionary.hasKey(serviceIdentifier);

    if (!bound && this.parent) {
      bound = this.parent.isBound(serviceIdentifier);
    }

    return bound;
  };

  _proto.isBoundNamed = function isBoundNamed(serviceIdentifier, named) {
    return this.isBoundTagged(serviceIdentifier, NAMED_TAG, named);
  } // Check if a binding with a complex constraint is available without throwing a error. Ancestors are also verified.
  ;

  _proto.isBoundTagged = function isBoundTagged(serviceIdentifier, key, value) {
    var bound = false; // verify if there are bindings available for serviceIdentifier on current binding dictionary

    if (this._bindingDictionary.hasKey(serviceIdentifier)) {
      var bindings = this._bindingDictionary.get(serviceIdentifier); // @ts-ignore


      var request = createMockRequest(this, serviceIdentifier, key, value);
      bound = bindings.some(function (b) {
        return b.constraint(request);
      });
    } // verify if there is a parent container that could solve the request


    if (!bound && this.parent) {
      bound = this.parent.isBoundTagged(serviceIdentifier, key, value);
    }

    return bound;
  };

  _proto.snapshot = function snapshot() {
    this._snapshots.push(ContainerSnapshot.of(this._bindingDictionary.clone(), this._middleware, this._activations.clone(), this._deactivations.clone(), this._moduleActivationStore.clone()));
  };

  _proto.restore = function restore() {
    var snapshot = this._snapshots.pop();

    if (snapshot === undefined) {
      throw new Error(NO_MORE_SNAPSHOTS_AVAILABLE);
    } // @ts-ignore


    this._bindingDictionary = snapshot.bindings;
    this._middleware = snapshot.middleware; // @ts-ignore

    this._activations = snapshot.activations; // @ts-ignore

    this._deactivations = snapshot.deactivations;
    this._middleware = snapshot.middleware;
    this._moduleActivationStore = snapshot.moduleActivationStore;
  } // @ts-ignore
  ;

  _proto.createChild = function createChild(containerOptions) {
    var child = new Container(containerOptions != null ? containerOptions : this.options); // @ts-ignore

    child.parent = this;
    return child;
  };

  _proto.applyMiddleware = function applyMiddleware() {
    var _this$_middleware;

    var initial = (_this$_middleware = this._middleware) != null ? _this$_middleware : this._planAndResolve();

    for (var _len5 = arguments.length, middlewares = new Array(_len5), _key6 = 0; _key6 < _len5; _key6++) {
      middlewares[_key6] = arguments[_key6];
    }

    this._middleware = middlewares.reduce(function (prev, curr) {
      return curr(prev);
    }, initial);
  };

  _proto.applyCustomMetadataReader = function applyCustomMetadataReader(metadataReader) {
    this._metadataReader = metadataReader;
  } // Resolves a dependency by its runtime identifier
  // The runtime identifier must be associated with only one binding
  // use getAll when the runtime identifier is associated with multiple bindings
  ;

  _proto.get = function get(serviceIdentifier) {
    var getArgs = this._getNotAllArgs(serviceIdentifier, false);

    return this._getButThrowIfAsync(getArgs); // return this._get<T>(
    //   false,
    //   false,
    //   TargetTypeEnum.Variable,
    //   serviceIdentifier
    // ) as T;
  };

  _proto.getAsync = /*#__PURE__*/function () {
    var _getAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee5(serviceIdentifier) {
      var getArgs;
      return _regeneratorRuntime().wrap(function _callee5$(_context5) {
        while (1) {
          switch (_context5.prev = _context5.next) {
            case 0:
              getArgs = this._getNotAllArgs(serviceIdentifier, false);
              return _context5.abrupt("return", this._get(getArgs));

            case 2:
            case "end":
              return _context5.stop();
          }
        }
      }, _callee5, this);
    }));

    function getAsync(_x3) {
      return _getAsync.apply(this, arguments);
    }

    return getAsync;
  }();

  _proto.getTagged = function getTagged(serviceIdentifier, key, value) {
    var getArgs = this._getNotAllArgs(serviceIdentifier, false, key, value);

    return this._getButThrowIfAsync(getArgs);
  };

  _proto.getTaggedAsync = /*#__PURE__*/function () {
    var _getTaggedAsync = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee6(serviceIdentifier, key, value) {
      var getArgs;
      return _regeneratorRuntime().wrap(function _callee6$(_context6) {
        while (1) {
          switch (_context6.prev = _context6.next) {
            case 0:
              getArgs = this._getNotAllArgs(serviceIdentifier, false, key, value);
              return _context6.abrupt("return", this._get(getArgs));

            case 2:
            case "end":
              return _context6.stop();
          }
        }
      }, _callee6, this);
    }));

    function getTaggedAsync(_x4, _x5, _x6) {
      return _getTaggedAsync.apply(this, arguments);
    }

    return getTaggedAsync;
  }();

  _proto.getNamed = function getNamed(serviceIdentifier, named) {
    return this.getTagged(serviceIdentifier, NAMED_TAG, named);
  };

  _proto.getNamedAsync = function getNamedAsync(serviceIdentifier, named) {
    return this.getTaggedAsync(serviceIdentifier, NAMED_TAG, named);
  } // Resolves a dependency by its runtime identifier
  // The runtime identifier can be associated with one or multiple bindings
  ;

  _proto.getAll = function getAll(serviceIdentifier) {
    var getArgs = this._getAllArgs(serviceIdentifier);

    return this._getButThrowIfAsync(getArgs);
  };

  _proto.getAllAsync = function getAllAsync(serviceIdentifier) {
    var getArgs = this._getAllArgs(serviceIdentifier);

    return this._getAll(getArgs);
  };

  _proto.getAllTagged = function getAllTagged(serviceIdentifier, key, value) {
    var getArgs = this._getNotAllArgs(serviceIdentifier, true, key, value);

    return this._getButThrowIfAsync(getArgs);
  };

  _proto.getAllTaggedAsync = function getAllTaggedAsync(serviceIdentifier, key, value) {
    var getArgs = this._getNotAllArgs(serviceIdentifier, true, key, value);

    return this._getAll(getArgs);
  };

  _proto.getAllNamed = function getAllNamed(serviceIdentifier, named) {
    return this.getAllTagged(serviceIdentifier, NAMED_TAG, named);
  };

  _proto.getAllNamedAsync = function getAllNamedAsync(serviceIdentifier, named) {
    return this.getAllTaggedAsync(serviceIdentifier, NAMED_TAG, named);
  };

  _proto.resolve = function resolve(constructorFunction) {
    var isBound = this.isBound(constructorFunction);

    if (!isBound) {
      this.bind(constructorFunction).toSelf();
    }

    var resolved = this.get(constructorFunction);

    if (!isBound) {
      this.unbind(constructorFunction);
    }

    return resolved;
  };

  _proto._preDestroy = function _preDestroy(constructor, instance) {
    if (reflection.hasMetadata(PRE_DESTROY, constructor)) {
      var data = reflection.getMetadata(PRE_DESTROY, constructor);
      return instance[data.value]();
    }
  } // @ts-ignore
  ;

  _proto._removeModuleHandlers = function _removeModuleHandlers(moduleId) {
    var moduleActivationsHandlers = this._moduleActivationStore.remove(moduleId);

    this._activations.removeIntersection(moduleActivationsHandlers.onActivations);

    this._deactivations.removeIntersection(moduleActivationsHandlers.onDeactivations);
  } // @ts-ignore
  ;

  _proto._removeModuleBindings = function _removeModuleBindings(moduleId) {
    return this._bindingDictionary.removeByCondition(function (binding) {
      return binding.moduleId === moduleId;
    });
  };

  _proto._deactivate = function _deactivate(binding, instance) {
    var _this4 = this;

    var constructor = Object.getPrototypeOf(instance).constructor;

    try {
      if (this._deactivations.hasKey(binding.serviceIdentifier)) {
        var result = this._deactivateContainer(instance, this._deactivations.get(binding.serviceIdentifier).values());

        if (isPromise(result)) {
          return this._handleDeactivationError(result.then(function () {
            return _this4._propagateContainerDeactivationThenBindingAndPreDestroyAsync(binding, instance, constructor);
          }), constructor);
        }
      }

      var propagateDeactivationResult = this._propagateContainerDeactivationThenBindingAndPreDestroy(binding, instance, constructor);

      if (isPromise(propagateDeactivationResult)) {
        return this._handleDeactivationError(propagateDeactivationResult, constructor);
      }
    } catch (ex) {
      throw new Error(ON_DEACTIVATION_ERROR(constructor.name, ex.message));
    }
  };

  _proto._handleDeactivationError = /*#__PURE__*/function () {
    var _handleDeactivationError2 = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee7(asyncResult, constructor) {
      return _regeneratorRuntime().wrap(function _callee7$(_context7) {
        while (1) {
          switch (_context7.prev = _context7.next) {
            case 0:
              _context7.prev = 0;
              _context7.next = 3;
              return asyncResult;

            case 3:
              _context7.next = 8;
              break;

            case 5:
              _context7.prev = 5;
              _context7.t0 = _context7["catch"](0);
              throw new Error(ON_DEACTIVATION_ERROR(constructor.name, _context7.t0.message));

            case 8:
            case "end":
              return _context7.stop();
          }
        }
      }, _callee7, null, [[0, 5]]);
    }));

    function _handleDeactivationError(_x7, _x8) {
      return _handleDeactivationError2.apply(this, arguments);
    }

    return _handleDeactivationError;
  }();

  _proto._deactivateContainer = function _deactivateContainer(instance, deactivationsIterator) {
    var _this5 = this;

    var deactivation = deactivationsIterator.next();

    while (deactivation.value) {
      var result = deactivation.value(instance);

      if (isPromise(result)) {
        return result.then(function () {
          return _this5._deactivateContainerAsync(instance, deactivationsIterator);
        });
      }

      deactivation = deactivationsIterator.next();
    }
  };

  _proto._deactivateContainerAsync = /*#__PURE__*/function () {
    var _deactivateContainerAsync2 = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee8(instance, deactivationsIterator) {
      var deactivation;
      return _regeneratorRuntime().wrap(function _callee8$(_context8) {
        while (1) {
          switch (_context8.prev = _context8.next) {
            case 0:
              deactivation = deactivationsIterator.next();

            case 1:
              if (!deactivation.value) {
                _context8.next = 7;
                break;
              }

              _context8.next = 4;
              return deactivation.value(instance);

            case 4:
              deactivation = deactivationsIterator.next();
              _context8.next = 1;
              break;

            case 7:
            case "end":
              return _context8.stop();
          }
        }
      }, _callee8);
    }));

    function _deactivateContainerAsync(_x9, _x10) {
      return _deactivateContainerAsync2.apply(this, arguments);
    }

    return _deactivateContainerAsync;
  }();

  _proto._getContainerModuleHelpersFactory = function _getContainerModuleHelpersFactory() {
    var _this6 = this;

    var setModuleId = function setModuleId(bindingToSyntax, moduleId) {
      bindingToSyntax._binding.moduleId = moduleId;
    };

    var getBindFunction = function getBindFunction(moduleId) {
      return function (serviceIdentifier) {
        var bindingToSyntax = _this6.bind(serviceIdentifier);

        setModuleId(bindingToSyntax, moduleId);
        return bindingToSyntax;
      };
    };

    var getUnbindFunction = function getUnbindFunction() {
      return function (serviceIdentifier) {
        return _this6.unbind(serviceIdentifier);
      };
    };

    var getUnbindAsyncFunction = function getUnbindAsyncFunction() {
      return function (serviceIdentifier) {
        return _this6.unbindAsync(serviceIdentifier);
      };
    };

    var getIsboundFunction = function getIsboundFunction(_moduleId) {
      return function (serviceIdentifier) {
        return _this6.isBound(serviceIdentifier);
      };
    };

    var getRebindFunction = function getRebindFunction(moduleId) {
      return function (serviceIdentifier) {
        var bindingToSyntax = _this6.rebind(serviceIdentifier);

        setModuleId(bindingToSyntax, moduleId);
        return bindingToSyntax;
      };
    };

    var getOnActivationFunction = function getOnActivationFunction(moduleId) {
      return function (serviceIdentifier, onActivation) {
        _this6._moduleActivationStore.addActivation(moduleId, serviceIdentifier, onActivation);

        _this6.onActivation(serviceIdentifier, onActivation);
      };
    };

    var getOnDeactivationFunction = function getOnDeactivationFunction(moduleId) {
      return function (serviceIdentifier, onDeactivation) {
        _this6._moduleActivationStore.addDeactivation(moduleId, serviceIdentifier, onDeactivation);

        _this6.onDeactivation(serviceIdentifier, onDeactivation);
      };
    };

    return function (mId) {
      return {
        bindFunction: getBindFunction(mId),
        isboundFunction: getIsboundFunction(mId),
        onActivationFunction: getOnActivationFunction(mId),
        onDeactivationFunction: getOnDeactivationFunction(mId),
        rebindFunction: getRebindFunction(mId),
        unbindFunction: getUnbindFunction(),
        unbindAsyncFunction: getUnbindAsyncFunction()
      };
    };
  };

  _proto._getAll = function _getAll(getArgs) {
    return Promise.all(this._get(getArgs));
  } // Prepares arguments required for resolution and
  // delegates resolution to _middleware if available
  // otherwise it delegates resolution to _planAndResolve
  ;

  _proto._get = function _get(getArgs) {
    var planAndResolveArgs = _extends({}, getArgs, {
      contextInterceptor: function contextInterceptor(context) {
        return context;
      },
      targetType: TargetTypeEnum.Variable
    });

    if (this._middleware) {
      var middlewareResult = this._middleware(planAndResolveArgs);

      if (middlewareResult === undefined || middlewareResult === null) {
        throw new Error(INVALID_MIDDLEWARE_RETURN);
      }

      return middlewareResult;
    }

    return this._planAndResolve()(planAndResolveArgs);
  };

  _proto._getButThrowIfAsync = function _getButThrowIfAsync(getArgs) {
    var result = this._get(getArgs);

    if (isPromiseOrContainsPromise(result)) {
      throw new Error(LAZY_IN_SYNC(getArgs.serviceIdentifier));
    }

    return result;
  };

  _proto._getAllArgs = function _getAllArgs(serviceIdentifier) {
    var getAllArgs = {
      avoidConstraints: true,
      isMultiInject: true,
      serviceIdentifier: serviceIdentifier
    };
    return getAllArgs;
  };

  _proto._getNotAllArgs = function _getNotAllArgs(serviceIdentifier, isMultiInject, key, value) {
    var getNotAllArgs = {
      avoidConstraints: false,
      isMultiInject: isMultiInject,
      serviceIdentifier: serviceIdentifier,
      key: key,
      value: value
    };
    return getNotAllArgs;
  } // Planner creates a plan and Resolver resolves a plan
  // one of the jobs of the Container is to links the Planner
  // with the Resolver and that is what this function is about
  ;

  _proto._planAndResolve = function _planAndResolve() {
    var _this7 = this;

    return function (args) {
      // create a plan
      var context = plan(_this7._metadataReader, // @ts-ignore
      _this7, args.isMultiInject, args.targetType, args.serviceIdentifier, args.key, args.value, args.avoidConstraints); // apply context interceptor

      context = args.contextInterceptor(context); // resolve plan

      var result = resolve(context);
      return result;
    };
  };

  _proto._deactivateIfSingleton = function _deactivateIfSingleton(binding) {
    var _this8 = this;

    if (!binding.activated) {
      return;
    }

    if (isPromise(binding.cache)) {
      return binding.cache.then(function (resolved) {
        return _this8._deactivate(binding, resolved);
      });
    }

    return this._deactivate(binding, binding.cache);
  };

  _proto._deactivateSingletons = function _deactivateSingletons(bindings) {
    for (var _iterator = _createForOfIteratorHelperLoose(bindings), _step; !(_step = _iterator()).done;) {
      var binding = _step.value;

      var result = this._deactivateIfSingleton(binding);

      if (isPromise(result)) {
        throw new Error(ASYNC_UNBIND_REQUIRED);
      }
    }
  };

  _proto._deactivateSingletonsAsync = /*#__PURE__*/function () {
    var _deactivateSingletonsAsync2 = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee9(bindings) {
      var _this9 = this;

      return _regeneratorRuntime().wrap(function _callee9$(_context9) {
        while (1) {
          switch (_context9.prev = _context9.next) {
            case 0:
              _context9.next = 2;
              return Promise.all(bindings.map(function (b) {
                return _this9._deactivateIfSingleton(b);
              }));

            case 2:
            case "end":
              return _context9.stop();
          }
        }
      }, _callee9);
    }));

    function _deactivateSingletonsAsync(_x11) {
      return _deactivateSingletonsAsync2.apply(this, arguments);
    }

    return _deactivateSingletonsAsync;
  }();

  _proto._propagateContainerDeactivationThenBindingAndPreDestroy = function _propagateContainerDeactivationThenBindingAndPreDestroy(binding, instance, constructor) {
    if (this.parent) {
      return this._deactivate.bind(this.parent)(binding, instance);
    } else {
      return this._bindingDeactivationAndPreDestroy(binding, instance, constructor);
    }
  };

  _proto._propagateContainerDeactivationThenBindingAndPreDestroyAsync = /*#__PURE__*/function () {
    var _propagateContainerDeactivationThenBindingAndPreDestroyAsync2 = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee10(binding, instance, constructor) {
      return _regeneratorRuntime().wrap(function _callee10$(_context10) {
        while (1) {
          switch (_context10.prev = _context10.next) {
            case 0:
              if (!this.parent) {
                _context10.next = 5;
                break;
              }

              _context10.next = 3;
              return this._deactivate.bind(this.parent)(binding, instance);

            case 3:
              _context10.next = 7;
              break;

            case 5:
              _context10.next = 7;
              return this._bindingDeactivationAndPreDestroyAsync(binding, instance, constructor);

            case 7:
            case "end":
              return _context10.stop();
          }
        }
      }, _callee10, this);
    }));

    function _propagateContainerDeactivationThenBindingAndPreDestroyAsync(_x12, _x13, _x14) {
      return _propagateContainerDeactivationThenBindingAndPreDestroyAsync2.apply(this, arguments);
    }

    return _propagateContainerDeactivationThenBindingAndPreDestroyAsync;
  }();

  _proto._removeServiceFromDictionary = function _removeServiceFromDictionary(serviceIdentifier) {
    try {
      this._bindingDictionary.remove(serviceIdentifier);
    } catch (e) {
      throw new Error(CANNOT_UNBIND + " " + getServiceIdentifierAsString(serviceIdentifier));
    }
  };

  _proto._bindingDeactivationAndPreDestroy = function _bindingDeactivationAndPreDestroy(binding, instance, constructor) {
    var _this10 = this;

    if (typeof binding.onDeactivation === 'function') {
      var result = binding.onDeactivation(instance);

      if (isPromise(result)) {
        return result.then(function () {
          return _this10._preDestroy(constructor, instance);
        });
      }
    }

    return this._preDestroy(constructor, instance);
  };

  _proto._bindingDeactivationAndPreDestroyAsync = /*#__PURE__*/function () {
    var _bindingDeactivationAndPreDestroyAsync2 = /*#__PURE__*/_asyncToGenerator( /*#__PURE__*/_regeneratorRuntime().mark(function _callee11(binding, instance, constructor) {
      return _regeneratorRuntime().wrap(function _callee11$(_context11) {
        while (1) {
          switch (_context11.prev = _context11.next) {
            case 0:
              if (!(typeof binding.onDeactivation === 'function')) {
                _context11.next = 3;
                break;
              }

              _context11.next = 3;
              return binding.onDeactivation(instance);

            case 3:
              _context11.next = 5;
              return this._preDestroy(constructor, instance);

            case 5:
            case "end":
              return _context11.stop();
          }
        }
      }, _callee11, this);
    }));

    function _bindingDeactivationAndPreDestroyAsync(_x15, _x16, _x17) {
      return _bindingDeactivationAndPreDestroyAsync2.apply(this, arguments);
    }

    return _bindingDeactivationAndPreDestroyAsync;
  }();

  return Container;
}();

var ContainerModule = function ContainerModule(registry) {
  this.id = void 0;
  this.registry = void 0;
  this.id = id();
  this.registry = registry;
};
var AsyncContainerModule = function AsyncContainerModule(registry) {
  this.id = void 0;
  this.registry = void 0;
  this.id = id();
  this.registry = registry;
};

exports.AsyncContainerModule = AsyncContainerModule;
exports.BindingScopeEnum = BindingScopeEnum;
exports.BindingTypeEnum = BindingTypeEnum;
exports.Container = Container;
exports.ContainerModule = ContainerModule;
exports.LazyServiceIdentifer = LazyServiceIdentifer;
exports.METADATA_KEY = metadata_keys;
exports.MetadataReader = MetadataReader;
exports.ModuleActivationStore = ModuleActivationStore;
exports.TargetTypeEnum = TargetTypeEnum;
exports.circularDependencyToException = circularDependencyToException;
exports.decorate = decorate;
exports.ensureFullyBound = ensureFullyBound;
exports.getFactoryDetails = getFactoryDetails;
exports.getFirstArrayDuplicate = getFirstArrayDuplicate;
exports.getFunctionName = getFunctionName;
exports.getServiceIdentifierAsString = getServiceIdentifierAsString;
exports.getSymbolDescription = getSymbolDescription;
exports.id = id;
exports.inject = inject;
exports.injectable = injectable;
exports.isClonable = isClonable;
exports.isPromise = isPromise;
exports.isPromiseOrContainsPromise = isPromiseOrContainsPromise;
exports.isStackOverflowExeption = isStackOverflowExeption;
exports.listMetadataForTarget = listMetadataForTarget;
exports.listRegisteredBindingsForServiceIdentifier = listRegisteredBindingsForServiceIdentifier;
exports.multiBindToService = multiBindToService;
exports.multiInject = multiInject;
exports.named = named;
exports.namedConstraint = namedConstraint;
exports.optional = optional;
exports.postConstruct = postConstruct;
exports.tagged = tagged;
exports.taggedConstraint = taggedConstraint;
exports.targetName = targetName;
exports.traverseAncerstors = traverseAncerstors;
exports.tryAndThrowErrorIfStackOverflow = tryAndThrowErrorIfStackOverflow;
exports.typeConstraint = typeConstraint;
exports.unmanaged = unmanaged;
//# sourceMappingURL=inversify-esm.cjs.development.js.map
